# Neural Modules

This repository contains:

- the architectures of the neural modules used in CLEVR, CLEVR-Dialog and MNIST-Dialog
- a flask server for using the neural modules through IRL primitives
- scripts for training and evaluating these modules on Hydra (or other clusters)

In this document, you will find the following sections:

- Setting up the Python environment for training the modules and running the IRL communication server.
- Downloading the raw datasets for CLEVR, CLEVR-Dialog and MNIST-Dialog
- Downloading pre-trained modules for CLEVR, CLEVR-Dialog and MNIST-Dialog.
- Running the flask server for hybrid procedural semantics with IRL.
- Generating or downloading the data that is required for training the neural modules from scratch.
- Using scripts for training the modules.
- Using scripts for evaluating the trained modules.
- Optional: other Makefile commands
- Optional: packages for (automatically) adhering to Python style guide using VSCode

## Setting up the Python environment

Anaconda is probably the fastest way to get up and running with Python environments. However, the full Anaconda toolkit installs 1500 Python packages and takes up 3GB of disk space. A good alternative, therefore, is Miniconda. Miniconda will only install the bare minimum to run Python. It is then up to you to manually install the packages that you _really_ need.

Download Miniconda from [this website](https://docs.conda.io/en/latest/miniconda.html) and follow the installation instructions. If you are prompted to add Miniconda to your `.bash_profile` or `.zshrc`, accept. For the changes to take effect and use `conda` from the command line, you must restart the Terminal. At the time of writing, the Miniconda website suggests to install Miniconda with Python 3.7.

The `environment.yml` file in this repository makes it easy to recreate the `neural-modules` conda environment that was used to develop this project. This file lists all necessary packages and their version numbers. You can use the Makefile to setup this environment.

1. To recreate the `neural-modules` conda environment that was used to develop this project, run:
   - `make create_environment`
2. If the environment was successfully created, you can activate it by running
   - `conda activate neural-modules`
   - You should see the prompt of your Terminal change from `(base)` to `(neural-modules)`.

## Downloading pre-trained models

All pretrained models are downloaded from the Babel Corpora SVN. From the root directory of the project, run the download script as follows:

```
bash scripts/download/pretrained_models.sh -u <your-svn-username>
```

## Downloading the CLEVR dataset

The CLEVR dataset can be downloaded through a bash script. From the root directory of the project, run the download script as follows:

```
bash scripts/download/clevr_raw.sh
```

Otherwise, the CLEVR dataset can also be downloaded [here](https://cs.stanford.edu/people/jcjohns/clevr/) from the original source.

## Downloading the MNIST-Dialog dataset

The MNIST-Dialog dataset can be downloaded through a bash script. From the root directory of the project, run the download script as follows:

```
bash scripts/download/mnist_dialog_raw.sh
```

Otherwise, the MNIST-Dialog dataset can also be downloaded [here](http://cvlab.postech.ac.kr/research/attmem/) from the original source.

## Running the Flask server

The Flask server makes use of a session when communicating with multiple instances, e.g. experiments using the neural modules that are running in parallel. Sessions are stored using Redis. To obtain Redis, run:

```
brew install redis
```

For starting the Flask server, use the file `server.py`. When starting the server, you need to specify which experiment you are running and which images will be made available for IRL.

```
python scripts/server/server.py
       --experiment [clevr, clevr_dialog, mnist_dialog]
       --images-dir <path-to-images-folder>
```

For the complete list of command line options, run:

```
python scripts/server/server.py --help
```

## Training modules from scratch

### Downloading training data

The training data that was used for the neural modules can be found on Babel Corpora under `neural-modules`. The following bash scripts can be used to obtain the data.

```
# Get the annotated data for training clevr modules
bash scripts/download/clevr_training.sh
# Get the annotated data for training clevr dialog modules
bash scripts/download/clevr_dialog_training.sh
# Get the annotated data for training mnist dialog modules
bash scripts/download/mnist_dialog_training.sh
```

%%% tot hier

### Obtain the enhanced CLEVR scenes

We have annotated the CLEVR JSON scene files with a mask for every object. The masks were obtained through a Mask R-CNN model pretrained on CLEVR-like images. The model and instructions on how to run it can be found [here](https://github.com/kexinyi/ns-vqa). The obtained masks were then matched to the symbolic CLEVR objects using the scripts `scripts/match_masks.py`.

To avoid re-running all of the above, you can simply contact Jens to obtain the training, validation and test set. These files are named `train.json`, `val.json` and `test.json` respectively and should be placed in the `data/` directory. The `clevr_config.json` file is also required, but should be already available in the `data/` directory.

The test set used in this project is different from the test set of the CLEVR dataset. This is due to the missing scene annotations from the CLEVR test set. These are however required to generate our training data. To compensate for this, we split the CLEVR training set in two parts of 60.000 scenes and 10.000 scenes respectively. The former serves as the training set and the latter as the test set. This split was made using random sampling as specified in the file `scripts/split_data.py`.

An additional consequence of this is that we need to filter out the correct 10.000 images from the CLEVR training images as the test set for this project. This can be done easily by iterating over `test.json`, each time reading the `image_filename` and copying this image to a separate directory. Contact Jens to obtain these images in a separate directory.

### Generating module specific datasets

The `generate_data.py` script can be used to generate training data for all modules at once. Run the following command from the root of the repository for each of the data files individually (train, val and test):

```bash
python scripts/generate_data.py --input_json data/train.json
```

By default, this script will read the config file from `data/clevr_config.json` and output the data to the `data/` directory. One can choose to skip particular modules by specifying their names using the `--ignore_modules` argument. When finished, the data folder should have the following structure:

```
data
| context
| count
| find_general
  | find_color
  | find_material
  | find_shape
  | find_size
| find_specific
  | find_blue
  | find_brown
  | find_cube
  | ...
  | find_yellow
| query_general
  | query_color
  | query_material
  | query_shape
  | query_size
| query_specific
  | query_blue
  | query_brown
  | query_cube
  | ...
  | query_yellow
| relate
| relate_specific
  | relate_behind
  | relate_front
  | relate_left
  | relate_right
```

### Running the training script

Finally, each module needs to be trained separately. For this, run the `scripts/train.py` script. Below, we provide an example of how to train a `query` module in the `specific` condition to query blue things:

```bash
python scripts/train.py
       --module query --condition specific --module_name query_blue \
       --train_images /path/to/CLEVR-v1.0/images/train \
       --val_images /path/to/CLEVR-v1.0/images/val \
       --train_scenes /data/query_specific/query_blue/train.json \
       --val_scenes /data/query_specific/query_blue/val.json
```

If you want more fine-grained control over the training script, take a look at `nmn/options.py`. This file specifies a.o. the command line options for the training script.

### Running the test script

If a module is done training, one can evaluate its performance on the held-out test set using the `scripts/test.py` script. This script will evaluate the module on the entire test set and append the name of the checkpoint and the accuracy to a csv file. The script can be ran as follows:

```bash
python scripts/test.py
       --checkpoint checkpoints/query_blue_000000000.pt \
       --test_images /path/to/test_images/ \
       --test_scenes  data/query_specific/query_blue/test.json \
       --output_csv output/test_accuracy.csv
```

### Running the evaluation script

If all modules have been trained and tested, one can evaluate the performance on the CLEVR questions and programs. This is done through the `scripts/evaluate.py` script. This script will read the CLEVR programs, assemble the modules as such, run them over the image and check the answer. This script can be called as follows:

```bash
python scripts/evaluate.py
       --question_json /path/to/CLEVR-v1.0/questions/CLEVR_train_questions.json \
       --images_dir /path/to/CLEVR-v1.0/images/train \
       --pretrained_models_dir pretrained_models \
       --condition general
```

In this example, the modules are evaluated on the CLEVR training set. The script assumes all pretrained models (i.e. `.pt` files) can be found in some `pretrained_models_dir`. Additionally, one should specify if one wants to test the general or specific condition. This will control which modules will be loaded.
