#!/usr/bin/env python

import numpy as np
from tqdm import tqdm

from nmn.datasets.generate.utils.masks import (
    combine_masks_for_extreme_relate,
    create_empty_mask,
)
from nmn.datasets.generate.utils.write import get_max_sample, write


def generate_extreme_relate_concept_dataset(args):
    relationships = args.config.relate
    for relationship in relationships:
        samples = []
        for scene in tqdm(args.dataset["scenes"]):
            objects = scene["objects"]
            relationships = scene["relationships"]
            # object with empty list (.index([])) has no other objects neighbors for that relationship
            # example: for the extreme left,
            # find the object of which there is
            # nothing to its left.
            idx = relationships[relationship].index([])
            target_mask = objects[idx]["mask"]
            sample = {"img_name": scene["image_filename"], "target": target_mask}
            samples.append(sample)

        samples = get_max_sample(args, samples)
        data = {
            "samples": samples,
        }
        write(
            args, data, module_name=f"extreme_relate_{relationship}",
        )

    # Separately handle the extreme_relate_middle !!
    samples = []
    for scene in tqdm(args.dataset["scenes"]):
        objects = scene["objects"]
        half_size = len(objects) / 2
        relationships = scene["relationships"]
        # Find the object(s) that are in the middle
        middle_idxs = []
        for i, _ in enumerate(objects):
            # an object is in the middle of the scene if it is in the middle of in all four spatial relationships
            # it is considered in the middle in a particular direction
            # for example given three objects (x,y,z)
            #   x is in the middle of the three objects => 3 // 2 -> 1 neighbour needed
            #   if it has 1 left neighbour
            #   if it has 1 right neighbour
            #   if it has 1 neighbour behind
            #   if it has 1 neighbour in front
            # Note:
            #   this definition was chosen by the original authors of the CLEVR_DIALOGUE dataset
            #   this scheme does introduce the possibility of multiple 'middle objects' being present.
            matches = np.sum(
                [
                    len(relationships[rel][i]) <= half_size
                    for rel in ["front", "behind", "left", "right"]
                ]
            )
            if matches == 4:
                middle_idxs.append(i)
        # Create masks for the middle object(s)
        if len(middle_idxs) == 0:
            target_mask = create_empty_mask(args)
        elif len(middle_idxs) == 1:
            target_mask = objects[middle_idxs[0]]["mask"]
        elif len(middle_idxs) > 1:
            middle_objects = [objects[i] for i in middle_idxs]
            target_mask = combine_masks_for_extreme_relate(args, middle_objects)
        # Create the sample
        sample = {"img_name": scene["image_filename"], "target": target_mask}
        samples.append(sample)

    samples = get_max_sample(args, samples)
    data = {
        "samples": samples,
    }
    write(args, data, module_name="extreme_relate_middle")
