#!/usr/bin/env python

import numpy as np
from tqdm import tqdm

from nmn.datasets.experiment import Experiment, UnsupportedExperimentError
from nmn.datasets.generate.utils.masks import calculate_immediate_relate_neighbour
from nmn.datasets.generate.utils.write import get_max_sample, write

########## QUERY CONCEPTS ##########


def get_query_concept_samples_clevr(scene, category, concept):
    samples = []
    objects = [obj for obj in scene["objects"] if "mask" in obj]
    for obj in objects:
        if obj[category] == concept:
            target_label = "yes"
        else:
            target_label = "no"
        sample = {
            "img_name": scene["image_filename"],
            "mask": obj["mask"],
            "target": target_label,
        }
        samples.append(sample)
    return samples


def get_query_concept_samples_mnist(scene, category, concept):
    samples = []
    digits = scene["digits"]
    for digit in digits:
        if digit[category] == concept:
            target_label = "yes"
        else:
            target_label = "no"
        sample = {
            "img_name": scene["img_name"],
            "mask": digit["mask"],
            "target": target_label,
        }
        samples.append(sample)
    return samples


def get_query_concept_samples(args, scene, category, concept):
    if args.experiment == Experiment.CLEVR:
        get_query_relate_concept_samples_clevr(args)
        get_query_immediate_relate_concept_samples_clevr(args)
        get_query_extreme_relate_concept_samples_clevr(args)
        get_query_extreme_relate_middle_samples_clevr(args)
        return get_query_concept_samples_clevr(scene, category, concept)
    elif args.experiment == Experiment.MNIST_DIALOG:
        return get_query_concept_samples_mnist(scene, category, concept)
    else:
        raise UnsupportedExperimentError(args.experiment)


def generate_query_concept_dataset(args):
    categories = args.config.categories
    for category in categories.keys():
        categories = categories[category]
        for concept in categories:
            samples = []
            for scene in tqdm(args.dataset["scenes"]):
                new_samples = get_query_concept_samples(args, scene, category, concept)
                samples.extend(new_samples)

            samples = get_max_sample(args, samples)
            data = {
                "samples": samples,
                "category": category,
            }
            write(args, data, module_name=f"query_{category}_{concept}")


########## QUERY RELATES ##########


def get_query_immediate_relate_concept_samples_clevr(args):
    relations = args.config.relate
    for relation in relations:
        samples = []
        for scene in tqdm(args.dataset["scenes"]):
            objects = scene["objects"]
            relationships = scene["relationships"]

            for current_idx in range(len(objects)):
                all_objects_idx = [i for i in range(len(objects))]
                all_objects_idx.remove(current_idx)
                neighbour_idx = calculate_immediate_relate_neighbour(
                    current_idx, relationships, relation
                )

                sample = {
                    "img_name": scene["image_filename"],
                    "mask": objects[current_idx]["mask"],
                    "target": "yes" if current_idx == neighbour_idx else "no",
                }
                samples.append(sample)

        samples = get_max_sample(args, samples)
        data = {"samples": samples}
        write(args, data, module_name=f"query_immediate_relate_{relation}")


def get_query_extreme_relate_concept_samples_clevr(args):
    relations = args.config.relate
    for relation in relations:
        samples = []
        for scene in tqdm(args.dataset["scenes"]):
            objects = scene["objects"]
            relationships = scene["relationships"]
            extreme_idx = relationships[relation].index([])

            for idx in range(len(objects)):
                sample = {
                    "img_name": scene["image_filename"],
                    "mask": objects[idx]["mask"],
                    "target": "yes" if idx == extreme_idx else "no",
                }
                samples.append(sample)

        samples = get_max_sample(args, samples)
        data = {"samples": samples}
        write(args, data, module_name=f"query_extreme_relate_{relation}")


def get_query_extreme_relate_middle_samples_clevr(args):
    # extreme_relate_middle case
    # Separately handle the extreme_relate_middle !!
    samples = []
    for scene in tqdm(args.dataset["scenes"]):
        objects = scene["objects"]
        half_size = len(objects) // 2
        relationships = scene["relationships"]
        # Find the object(s) that are in the middle
        middle_idxs = []
        for i, _ in enumerate(objects):
            # An object is in the middle of the scene if it is in the middle of in all four spatial relationships
            #   it is considered in the middle in a particular direction
            # for example given three objects (x,y,z)
            #   x is in the middle of the three objects => 3 // 2 -> 1 neighbour needed
            #   if it has 1 left neighbour
            #   if it has 1 right neighbour
            #   if it has 1 neighbour behind
            #   if it has 1 neighbour in front
            # Note:
            #   this definition was chosen by the original authors of the CLEVR_DIALOGUE dataset
            #   this scheme does introduce the possibility of multiple 'middle objects' being present.
            #   our implementation creates a sample for all 'middle objects'
            matches = np.sum(
                [
                    len(relationships[rel][i]) <= half_size
                    for rel in ["front", "behind", "left", "right"]
                ]
            )
            if matches == 4:
                middle_idxs.append(i)

        for idx in range(len(objects)):
            sample = {
                "img_name": scene["image_filename"],
                "mask": objects[idx]["mask"],
                "target": "yes" if idx in middle_idxs else "no",
            }
            samples.append(sample)

    samples = get_max_sample(args, samples)
    data = {"samples": samples}
    write(args, data, module_name="query_extreme_relate_middle")


def get_query_relate_concept_samples_clevr(args):
    relations = args.config.relate
    for relation in relations:
        samples = []
        for scene in tqdm(args.dataset["scenes"]):
            objects = scene["objects"]
            relationships = scene["relationships"]
            neighbor_idxs = relationships[relation]
            for current_obj_idx, neighbors_with_idx in enumerate(neighbor_idxs, 0):
                set_all_objects = set([i for i in range(len(objects))])
                set_all_objects.remove(current_obj_idx)
                set_neighbors = set(neighbors_with_idx)
                set_non_neighors = set_all_objects.difference(set_neighbors)
                current_mask = objects[current_obj_idx]["mask"]

                for idx in set_neighbors:
                    # neighbor_mask is {relation} of current_mask
                    neighbor_mask = objects[idx]["mask"]
                    sample = {
                        "img_name": scene["image_filename"],
                        "mask1": current_mask,
                        "mask2": neighbor_mask,
                        "target": "yes",
                    }
                    samples.append(sample)

                for idx in set_non_neighors:
                    # neighbor_mask is not {relation} of current_mask
                    neighbor_mask = objects[idx]["mask"]
                    sample = {
                        "img_name": scene["image_filename"],
                        "mask1": current_mask,
                        "mask2": neighbor_mask,
                        "target": "no",
                    }
                    samples.append(sample)

            samples = get_max_sample(args, samples)
            data = {"samples": samples}
            write(args, data, module_name=f"query_relate_{relation}")
