import numpy as np
from pycocotools import mask


def create_empty_mask(args):
    """
    Creates an empty pycocotools mask in RLE format
    """
    # height and width standardized accross entire dataset
    height = args.config.img_height
    width = args.config.img_width

    # matrix of data-type uint8 as required by pycocotools
    empty_matrix = np.zeros((height, width), dtype=np.uint8)
    # pycocotools requires the array to be F(ortran)-CONTIGUOUS
    empty_matrix = np.asfortranarray(empty_matrix)

    empty_mask = mask.encode(empty_matrix)
    # required to store as json
    empty_mask["counts"] = empty_mask["counts"].decode("utf-8")
    return empty_mask


def create_full_mask(args):
    """
    Creates a completely covered mask in RLE format
    """
    # height and width standardized accross entire dataset
    height = args.config.img_height
    width = args.config.img_width

    # matrix of data-type uint8 as required by pycocotools
    matrix = np.ones((height, width), dtype=np.uint8)
    # pycocotools requires the array to be F(ortran)-CONTIGUOUS
    matrix = np.asfortranarray(matrix)

    full_mask = mask.encode(matrix)
    # required to store as json
    full_mask["counts"] = full_mask["counts"].decode("utf-8")
    return full_mask


def create_background_mask(objects):
    context_mask = create_context_mask(objects)
    raw = mask.decode(context_mask)
    raw ^= 1  # flip all binary bits
    input_mask = mask.encode(raw)
    # required to store as json
    input_mask["counts"] = input_mask["counts"].decode("utf-8")
    return input_mask


def create_context_mask(objects):
    input_masks = []
    for obj in objects:
        input_masks.append(obj["mask"])
    context_mask = mask.merge(input_masks)
    # required to store as json
    context_mask["counts"] = context_mask["counts"].decode("utf-8")
    return context_mask


def combine_masks_for_find(args, objects, category, concept):
    input_masks = [create_empty_mask(args)]  # add empty mask to deal with 0 results
    output_masks = [create_empty_mask(args)]  # add empty mask to deal with 0 results

    for obj in objects:
        input_masks.append(obj["mask"])
        if obj[category] == concept:
            output_masks.append(obj["mask"])

    input_mask = mask.merge(input_masks)
    output_mask = mask.merge(output_masks)
    # required to store as json
    input_mask["counts"] = input_mask["counts"].decode("utf-8")
    output_mask["counts"] = output_mask["counts"].decode("utf-8")

    return input_mask, output_mask


def combine_masks_for_relate(args, objects, object, relationships):
    input_mask = object["mask"]
    output_masks = [create_empty_mask(args)]  # add empty mask to deal with 0 results

    for idx in relationships:
        if "mask" in objects[idx]:
            output_masks.append(objects[idx]["mask"])

    output_mask = mask.merge(output_masks)
    # required to store as json
    output_mask["counts"] = output_mask["counts"].decode("utf-8")
    return input_mask, output_mask


def combine_masks_for_extreme_relate(args, objects):
    target_mask = mask.merge([obj["mask"] for obj in objects])
    target_mask["counts"] = target_mask["counts"].decode("utf-8")
    return target_mask


def calculate_immediate_relate_neighbour(idx, relationships, relationship):
    right_relationships = {}
    for id, obj in enumerate(relationships[relationship]):
        right_relationships[id] = len(obj)
    y_pos_list = sorted(right_relationships, key=right_relationships.get)
    y_pos_list.reverse()
    pos_in_y_pos_list = y_pos_list.index(idx)

    if pos_in_y_pos_list + 1 < len(y_pos_list):
        # index of the object <relationship> the first object
        target_index = y_pos_list[pos_in_y_pos_list + 1]
        return target_index
    else:
        return None


def create_immediate_relate_mask(args, idx, objects, relationships, relationship):
    neighbor_idx = calculate_immediate_relate_neighbour(idx, relationships, relationship)
    if neighbor_idx:
        return objects[neighbor_idx]["mask"]
    else:
        return create_empty_mask(args)


def create_mask_of_closest_object(args, object, objects, consider_idxs, relationship):
    if len(consider_idxs) > 0:
        closest_object = None
        smallest_difference = None
        object_x = object["pixel_coords"][0]
        object_y = object["pixel_coords"][1]
        # for these indexes, we are sure that the relationship holds
        # e.g. for 'left', all consider_idxs are left of 'object'
        for idx in consider_idxs:
            other = objects[idx]
            other_x = other["pixel_coords"][0]
            other_y = other["pixel_coords"][1]
            if relationship == "left":
                diff = abs(object_x - other_x)
            elif relationship == "right":
                diff = abs(object_x - other_x)
            elif relationship == "front":
                diff = abs(object_y - other_y)
            elif relationship == "behind":
                diff = abs(object_y - other_y)

            if closest_object is None:
                closest_object = other
                smallest_difference = diff
            elif diff < smallest_difference:
                closest_object = other
                smallest_difference = diff
        return closest_object["mask"]
    else:
        return create_empty_mask(args)


def is_in_mask(x, y, w, a, b):
    """
    Verifies if a set of pixels is within the boundaries of the bounding box of a digit in the grid.

    This method is originates (and is specific to) from the MNIST codebase.
    TODO should be abstracted and moved.

    Args:
        x,y: integers, pixel position for which the check is done
        w: integer, size of the bounding box starting from pixel position (a,b)
        a,b: integers: pixel position of the top left corner of the bounding box

    Returns:
        A boolean
    """
    return (a >= x) and (b >= y) and (a < x + w) and (b < y + w)


def create_mnist_mask_by_index(idx):
    """Creates and returns an mask at idx of the grid

    This method is originates (and is specific to) from the MNIST codebase.
    TODO should be abstracted and moved.

    Grid layout:

        0, 1, 2, 3
        4, 5, 6, 7
        8, 9, 10, 11
        12, 13, 14, 15

    Args:
        idx: integer, represents the index in the grid layout for which a mask is created

    Returns:
        A dictionary representing an binary mask stored in RLE format (compressed) and its dims
    """
    # height and width standardized across entire dataset
    height = 112
    width = 112
    gridsize = 4
    x, y = idx // gridsize, idx % gridsize
    x, y = x * (height / gridsize), y * (width / gridsize)

    # matrix of data-type uint8 as required by pycocotools
    matrix = np.zeros((height, width), dtype=np.uint8)

    # set every pixel within the given grid position to 1, rest to 0
    for a in range(height):
        for b in range(width):
            if is_in_mask(x, y, width / gridsize, a, b):
                matrix[a][b] = 1

    # pycocotools requires the array to be F(ortran)-CONTIGUOUS
    matrix = np.asfortranarray(matrix)
    RLE_mask = mask.encode(matrix)
    RLE_mask["counts"] = RLE_mask["counts"].decode("utf-8")
    return RLE_mask


def calculate_mid_mask(RLE_mask):
    """Calculates given a binary mask the middle point of the mask."""
    binary_mask = mask.decode(RLE_mask)
    y_nonzero, x_nonzero = np.nonzero(binary_mask)
    x_nonzero_arr = list(set(x_nonzero))
    x_mid = x_nonzero_arr[int(len(x_nonzero_arr) / 2)]
    y_nonzero_arr = list(set(y_nonzero))
    y_mid = y_nonzero_arr[int(len(y_nonzero_arr) / 2)]
    return [x_mid, y_mid]  # TODO fix should be a tuple
