#!/usr/bin/env pythons

import random

from torch.utils.data import DataLoader, Subset, WeightedRandomSampler

from nmn.datasets.load.experiment_dataset import get_dataset
from nmn.datasets.load.modules.context import ContextModuleDataset
from nmn.datasets.load.modules.count import CountModuleDataset
from nmn.datasets.load.modules.extreme_relate_concept import ExtremeRelateConceptModuleDataset
from nmn.datasets.load.modules.find_category import FindCategoryModuleDataset
from nmn.datasets.load.modules.find_concept import FindConceptModuleDataset
from nmn.datasets.load.modules.immediate_relate_category import ImmediateRelateCategoryModuleDataset
from nmn.datasets.load.modules.immediate_relate_concept import ImmediateRelateConceptModuleDataset
from nmn.datasets.load.modules.query_category import QueryCategoryModuleDataset
from nmn.datasets.load.modules.query_concept import QueryConceptModuleDataset
from nmn.datasets.load.modules.relate_category import RelateCategoryModuleDataset
from nmn.datasets.load.modules.relate_concept import RelateConceptModuleDataset
from nmn.datasets.load.utils.weights import get_class_weights
from nmn.modules.utils.module_type import ModuleType, UnsupportedModuleTypeError


def get_dataloader(**kwargs):
    dataset_kwargs = {
        "module_info": kwargs.get("module_info"),
        "images": kwargs.get("images"),
        "scenes": kwargs.get("scenes"),
        "config": kwargs.get("config"),
        "device": kwargs.get("device"),
    }
    ds = get_module_dataset(**dataset_kwargs)

    # dataloader kwargs
    dataloader_kwargs = {
        "batch_size": kwargs.get("batch_size"),
        "shuffle": kwargs.get("shuffle"),
        "num_workers": kwargs.get("num_workers"),
    }
    # Get the class weights using the entire dataset
    balance_classes = kwargs.get("balance_classes", False)
    class_weights = None
    if balance_classes:
        class_weights = get_class_weights(ds)
    # If max_samples is specified, take a subset
    max_samples = kwargs.get("max_samples")
    if max_samples is not None and max_samples <= len(ds):
        # take a random sample of size max_samples using the indices
        all_indices = list(range(0, len(ds)))
        subset = random.sample(all_indices, max_samples)
        ds = Subset(ds, subset)
    # If class weights exist, use them on the Dataset/Subset
    if class_weights is not None:
        if type(ds) is Subset:
            indices = ds.indices
            all_samples = [ds.dataset.samples[idx] for idx in indices]
        else:
            all_samples = ds.samples
        all_targets = [sample["target"] for sample in all_samples]
        sample_weights = [class_weights[t] for t in all_targets]
        weighted_sampler = WeightedRandomSampler(sample_weights, len(ds))
        loader = DataLoader(dataset=ds, sampler=weighted_sampler, **dataloader_kwargs)
    else:
        loader = DataLoader(dataset=ds, **dataloader_kwargs)
    return loader


def get_module_dataset(**kwargs):
    module_type = kwargs.get("module_info").type
    module_specificity = (
        kwargs.get("module_info").specificity
        if "specificity" in kwargs.get("module_info")
        else None
    )
    if module_type == ModuleType.CONTEXT:
        dataset = ContextModuleDataset(**kwargs)
    elif module_type == ModuleType.COUNT:
        dataset = CountModuleDataset(**kwargs)
    elif module_type == ModuleType.FIND:
        if module_specificity == "category":
            dataset = FindCategoryModuleDataset(**kwargs)
        elif module_specificity == "concept":
            dataset = FindConceptModuleDataset(**kwargs)
    elif module_type == ModuleType.QUERY:
        if module_specificity == "category":
            dataset = QueryCategoryModuleDataset(**kwargs)
        elif module_specificity == "concept":
            dataset = QueryConceptModuleDataset(**kwargs)
    elif module_type == ModuleType.RELATE:
        if module_specificity == "category":
            dataset = RelateCategoryModuleDataset(**kwargs)
        elif module_specificity == "concept":
            dataset = RelateConceptModuleDataset(**kwargs)
    elif module_type == ModuleType.EXTREME_RELATE:
        if module_specificity == "category":
            raise ValueError(
                "The dataset for module {module_type}_{module_specificity} has not yet been implemented."
            )
        elif module_specificity == "concept":
            dataset = ExtremeRelateConceptModuleDataset(**kwargs)
    elif module_type == ModuleType.IMMEDIATE_RELATE:
        if module_specificity == "category":
            dataset = ImmediateRelateCategoryModuleDataset(**kwargs)
        elif module_specificity == "concept":
            dataset = ImmediateRelateConceptModuleDataset(**kwargs)
    else:
        raise UnsupportedModuleTypeError(module_type)
    return dataset


def get_evaluation_dataloader(**kwargs):
    experiment, questions, images_dir, device = (
        kwargs.pop("experiment"),
        kwargs.pop("questions"),
        kwargs.pop("images_dir"),
        kwargs.pop("device"),
    )
    ds = get_dataset(experiment, questions, images_dir, device)
    max_samples = kwargs.pop("max_samples")
    if max_samples is not None:
        all_indices = list(range(0, len(ds)))
        if kwargs["shuffle"] == 0:
            subset = all_indices[:max_samples]
        else:
            subset = random.sample(all_indices, max_samples)
        ds = Subset(ds, subset)
    loader = DataLoader(dataset=ds, **kwargs)
    return loader
