#!/usr/bin/env python

import numpy as np
import torch
import torch.nn as nn
from pycocotools import mask as pycoco
from skimage.util import random_noise


class OneHotEncoding(nn.Module):
    """
    OneHotEncoding class .
    returns a one-hot encoding for any word in the vocabulary
    of the experiment based on the pytorch nn.Embedding module
    :param vocab: the vocabulary it needs to encode
    """

    def __init__(self, config):
        super(OneHotEncoding, self).__init__()

        self.vocab = config.vocab
        self.word_to_ix = {v: i for i, v in enumerate(self.vocab)}
        embedding = torch.eye(self.encoding_size)
        self.embedding = nn.Embedding.from_pretrained(embedding, freeze=True)

    def forward(self, word):
        lookup_tensor = torch.tensor([self.word_to_ix[word]], dtype=torch.long)
        return self.embedding.forward(lookup_tensor)


class GloveEncoding(nn.Module):
    """
    Glove-based encoding class.
    returns the Glove encoding for any word in the vocabulary
    of the experiment based on the pytorch nn.Embedding module
    :param vocab: the vocabulary it needs to encode
    """

    def __init__(self, config):
        super(GloveEncoding, self).__init__()
        self.vocab = config.vocab
        self.word_to_ix = {v: i for i, v in enumerate(self.vocab)}
        embedding = torch.from_numpy(config.weights)
        self.embedding = nn.Embedding.from_pretrained(embedding, freeze=True)

    def forward(self, word):
        lookup_tensor = torch.tensor([self.word_to_ix[word]], dtype=torch.long)
        return self.embedding.forward(lookup_tensor)


class ImageToTensor(object):
    """Convert image to a tensor."""

    def __init__(self, device):
        super(ImageToTensor, self).__init__()
        self.device = device

    def __call__(self, image):
        image = image[:, :, :3]
        image = image.transpose((2, 0, 1))
        image = image / 255
        image = torch.from_numpy(image)
        image = image.to(self.device, dtype=torch.float)
        return image


class MaskToTensor(object):
    """Convert mask to a tensor."""

    def __init__(self, device):
        super(MaskToTensor, self).__init__()
        self.device = device

    def __call__(self, mask, noise=0.0):
        # pycoco mask to numpy matrix
        mask = pycoco.decode(mask)
        # reshape: H X W -> 1 X H X W
        mask = np.expand_dims(mask, axis=0)
        # possibly add noise
        if noise != 0.0:
            mask = random_noise(mask, mode="gaussian", clip=True, mean=0, var=noise)
        # convert np arrays to tensors
        mask = torch.from_numpy(mask)
        mask = mask.to(self.device, dtype=torch.float)
        return mask


class OneHotToTensor(object):
    """Convert text to one hot tensor"""

    def __init__(self, device):
        super(OneHotToTensor, self).__init__()
        self.device = device

    def __call__(self, embedding):
        embedding = torch.from_numpy(embedding)
        embedding = embedding.to(self.device, dtype=torch.long)
        return embedding


class GloveToTensor(object):
    """Convert text to glove tensor"""

    def __init__(self, config, device):
        self.glove = GloveEncoding(config)
        self.device = device

    def __call__(self, word):
        embedding = self.glove(word)
        embedding = embedding.to(self.device, dtype=torch.float)
        return embedding
