#!/usr/bin/env python

import logging
from collections import defaultdict

import numpy as np


class ExperimentEvaluator(object):
    """
    The ExperimentEvaluator controls the evaluation process of complete
    networks of modules.
    """

    def __init__(self, program_executor, data_loader, mode):
        self.program_executor = program_executor
        self.data_loader = data_loader
        self.mode = mode

    def individual_performance(self):
        """Performance of the modules during evaluation of the given task.

        This method originates from the CLEVR VQA codebase.
        """
        logging.info("===== Start evaluation =====")
        correct_per_type = defaultdict(list)
        for i, batch in enumerate(self.data_loader):
            logging.info("============================================")
            logging.info(f"Sample {i}")
            image, program, answer = batch
            success = self.program_executor(image, program, answer, i)
            correct = 1 if success else 0
            last_func = program[-1]["function"]
            correct_per_type["all"].append(correct)
            correct_per_type[last_func].append(correct)
            del image, program, answer, batch
        logging.info("===== Finished evaluation =====")
        for q_type, vals in sorted(correct_per_type.items()):
            vals = np.asarray(vals)
            logging.info(
                f"{q_type} {vals.sum()} / {vals.shape[0]} = {100.0 * vals.mean():.2f}"
            )

    def grouped_performance(self):
        """Performance of the modules during evaluation of the given task.

        In this function, the performance of certain modules are grouped.
        For example, the 'equal_integer', 'less_than', 'greater_than' modules
            are grouped into the 'compare_integer' group.

        This method originates from the MNIST codebase.
        """
        logging.info("===== Start evaluation =====")
        num_questions = 0
        num_correct = 0
        total_per_type = {}
        correct_per_type = {}
        for i, batch in enumerate(self.data_loader):
            logging.info("=" * 50)
            logging.info("Sample", i)
            image, program, answer = batch
            success = self.program_executor(image, program, answer, i)
            last_func = program[-1]["function"]
            if last_func in ["equal_integer", "less_than", "greater_than"]:
                question_type = "compare_integer"
            elif "query" in last_func:
                question_type = "query"
            elif "equal" in last_func:
                question_type = "compare"
            else:
                question_type = last_func
            if success:
                num_correct += 1
                if question_type in correct_per_type:
                    correct_per_type[question_type] += 1
                else:
                    correct_per_type[question_type] = 1
            num_questions += 1
            if question_type in total_per_type:
                total_per_type[question_type] += 1
            else:
                total_per_type[question_type] = 1
            del image, program, answer, batch
        logging.info("Overal accuracy:", num_correct / float(num_questions))
        logging.info("Accuracy per question type:")
        for key in total_per_type:
            accuracy = correct_per_type[key] / float(total_per_type[key])
            logging.info(f"  - {key}: {accuracy}")
        logging.info("===== Finished evaluation =====")

    def evaluate(self):
        if self.mode == "individual":
            self.individual_performance()
        elif self.mode == "grouped":
            self.grouped_performance()
        else:
            raise ValueError(f"Unexpected evaluation_mode: {self.mode}.")
