#!/usr/bin/env python

import csv
import logging
import os
import pickle

import numpy as np
import torch
from tqdm import tqdm


class ModuleEvaluator(object):
    """
    The Tester specifies the evaluation loop for a trained modules.
    It gets data from the test_loader and passes it to the model,
    which computes the loss and the accuracy. At the end, the loss
    and accuracy of the module are exported to a file.
    """

    def __init__(self, opt, model, test_loader):
        self.model = model
        self.test_loader = test_loader
        self.output_csv = opt.output_csv
        self.checkpoint_file = opt.checkpoint
        self.save_num_batches = opt.save_num_batches
        self.predictions_pkl = opt.predictions_pkl
        self.device = opt.device

    def run(self):
        logging.info("===== Start testing =====")
        # check the predictions_pkl if necessary
        if self.save_num_batches > 0:
            name, ext = os.path.splitext(self.predictions_pkl)
            if ext != ".pkl":
                raise ("Expected a .pkl file, got %s" % self.predictions_pkl)
        # start testing
        self.model.set_mode("eval")
        losses = []
        accuracies = []
        t = 0
        export_num_batches = self.save_num_batches
        export = []
        for batch in tqdm(self.test_loader):
            t += 1
            source, target = batch
            target = target.to(self.device)
            with torch.no_grad():
                loss, accuracy, prediction = self.model.forward(
                    source, target, mode="test"
                )
            losses.append(loss.item())
            accuracies.append(accuracy)

            if export_num_batches > 0:
                # export predictions of batch
                batch_size = target.shape[0]
                for b in range(batch_size):
                    entry = {}
                    for key, value in source.items():
                        if type(value[b]) is torch.Tensor:
                            entry[key] = value[b].detach().numpy()
                        else:
                            entry[key] = value[b]
                    if type(target[b]) is torch.Tensor:
                        entry["target"] = target[b].detach().numpy()
                    else:
                        entry["target"] = target[b]
                    if type(prediction[b]) is torch.Tensor:
                        entry["prediction"] = prediction[b].detach().numpy()
                    else:
                        entry["prediction"] = prediction[b]
                    export.append(entry)
                # decrement counter
                export_num_batches -= 1

        avg_loss = np.mean(losses).item()
        avg_acc = np.mean(accuracies).item()

        if self.save_num_batches > 0:
            with open(self.predictions_pkl, "wb") as f:
                pickle.dump(export, f)

        fieldnames = ["checkpoint", "loss", "accuracy"]
        if not os.path.isfile(self.output_csv):
            with open(self.output_csv, "w") as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()

        with open(self.output_csv, "a") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writerow(
                {
                    "checkpoint": os.path.basename(self.checkpoint_file),
                    "loss": avg_loss,
                    "accuracy": avg_acc,
                }
            )
        logging.info(f"Test set accuracy is {avg_acc}")
        logging.info("===== Finished testing =====")
