#!/usr/bin/env python

import numpy as np
import torch.nn as nn
from sklearn.metrics import jaccard_score as iou

from nmn.modules.utils.module_type import ModuleType


class MultiClassAccuracy(nn.Module):
    def __init__(self):
        super(MultiClassAccuracy, self).__init__()

    def forward(self, probs, target):
        num_correct = (probs.argmax(dim=1) == target).sum().float()
        accuracy = num_correct / float(target.size(0))
        return accuracy.detach().cpu().numpy()


class IOUAccuracy(nn.Module):
    """Calculates the IOU accuracy of two masks."""

    def __init__(self):
        super(IOUAccuracy, self).__init__()

    def forward(self, pred, target):
        # Detach the tensors from the GPU
        pred = pred.detach().cpu().numpy()
        target = target.detach().cpu().numpy()
        # final step was a sigmoid,
        # thus need to discretize the prediction
        mask = np.where(pred <= 0.5, 0, 1)
        # Mask and target have the following shape:
        # [BATCH, CHANNELS, HEIGHT, WIDTH]
        # For IOU, we need to compare single-dim arrays
        # so we flatten them
        batch_size = target.shape[0]
        t = target.reshape(batch_size, -1)
        m = mask.reshape(batch_size, -1)
        accs = [iou(t[i], m[i], average="macro") for i in range(batch_size)]
        return np.mean(accs)


def get_evaluation_fn(module_type):
    if module_type == ModuleType.CONTEXT:
        return IOUAccuracy()
    elif module_type == ModuleType.COUNT:
        return MultiClassAccuracy()
    elif module_type == ModuleType.FIND:
        return IOUAccuracy()
    elif module_type == ModuleType.QUERY:
        return MultiClassAccuracy()
    elif module_type == ModuleType.RELATE:
        return IOUAccuracy()
    elif module_type == ModuleType.EXTREME_RELATE:
        return IOUAccuracy()
    elif module_type == ModuleType.IMMEDIATE_RELATE:
        return IOUAccuracy()
