#!/usr/bin/env python

import logging
import os
import pickle

import torch
import torch.nn as nn


class ProgramExecutor(nn.Module):
    """
    The ProgramExecutor class handles the processing of procedural
    programs with neural-modules. It handles the passing of data
    from one module to the next and runs an entire program.

    Only relevant for CLEVR, not for CLEVR_Dialog or MNIST_Dialog.
    """

    def __init__(
        self,
        modules,
        specificity,
        config,
        export_trace_pkl,
        export_answers,
        use_clean_count_masks,
        use_clean_query_masks,
        count_clean_mask_threshold,
        query_clean_mask_threshold,
    ):
        super(ProgramExecutor, self).__init__()
        self.modules = modules
        self.specificity = specificity
        self.config = config
        self.outputs = {}
        self.export_trace_pkl = export_trace_pkl
        self.export_answers = export_answers
        self.use_clean_count_masks = use_clean_count_masks
        self.use_clean_query_masks = use_clean_query_masks
        self.count_clean_mask_threshold = count_clean_mask_threshold
        self.query_clean_mask_threshold = query_clean_mask_threshold
        # if the answers file exists, delete it
        if export_answers is not None and os.path.exists(export_answers):
            os.remove(export_answers)

    def make_clean_mask(self, mask, threshold):
        if mask.is_cuda:
            return torch.where(
                mask <= threshold, torch.Tensor([0]).cuda(), torch.Tensor([1]).cuda()
            )
        else:
            return torch.where(mask <= threshold, torch.Tensor([0]), torch.Tensor([1]))

    def run_program(self, program, image):
        for i, entry in enumerate(program):
            function = entry["function"]
            if len(entry["value_inputs"]) > 0:
                value_input = entry["value_inputs"][0]
            inputs = entry["inputs"]
            # get context
            if function == "scene":
                module = self.modules["context"]
                attn = module({"image": image})
                self.outputs[i] = attn
            # filter
            elif "filter" in function:
                module = self.modules["filter"]
                attn = module(
                    {
                        "image": image,
                        "mask": self.outputs[inputs[0]],
                        "category": function.split("_")[-1],
                        "concept": value_input,
                    }
                )
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = attn
            # query
            elif "query" in function:
                module = self.modules["query"]
                category = function.split("_")[-1]
                mask = self.outputs[inputs[0]]
                if self.use_clean_query_masks:
                    mask = self.make_clean_mask(mask, self.query_clean_mask_threshold)
                index, _ = module(
                    {
                        "image": image,
                        "mask": mask,
                        "category": category,
                        "options": self.config.categories[category],
                    }
                )
                label = self.config.categories[category][index]
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = label
            # equal
            elif "equal" in function:
                module = self.modules["equal"]
                category1 = self.outputs[inputs[0]]
                category2 = self.outputs[inputs[1]]
                bool = module(category1, category2)
                # self.outputs[inputs[0]] = None # free up memory
                # self.outputs[inputs[1]] = None # free up memory
                self.outputs[i] = "yes" if bool else "no"
            # same
            elif "same" in function:
                module = self.modules["same"]
                category = function.split("_")[-1]
                options = self.config.categories[category]
                mask = self.outputs[inputs[0]]
                if self.use_clean_query_masks:
                    mask = self.make_clean_mask(mask, self.query_clean_mask_threshold)
                attn = module(
                    {
                        "image": image,
                        "mask": mask,
                        "category": category,
                        "options": options,
                    }
                )
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = attn
            # relate
            elif function == "relate":
                module = self.modules["relate"]
                attn = module(
                    {
                        "image": image,
                        "mask": self.outputs[inputs[0]],
                        "relationship": value_input,
                    }
                )
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = attn
            # intersection
            elif function == "intersect":
                module = self.modules["intersect"]
                attn = module(self.outputs[inputs[0]], self.outputs[inputs[1]])
                # self.outputs[inputs[0]] = None # free up memory
                # self.outputs[inputs[1]] = None # free up memory
                self.outputs[i] = attn
            # union
            elif function == "union":
                module = self.modules["union"]
                attn = module(self.outputs[inputs[0]], self.outputs[inputs[1]])
                # self.outputs[inputs[0]] = None # free up memory
                # self.outputs[inputs[1]] = None # free up memory
                self.outputs[i] = attn
            # greater than
            elif function == "greater_than":
                module = self.modules["greater_than"]
                bool = module(self.outputs[inputs[0]], self.outputs[inputs[1]])
                # self.outputs[inputs[0]] = None # free up memory
                # self.outputs[inputs[1]] = None # free up memory
                self.outputs[i] = "yes" if bool else "no"
            # less than
            elif function == "less_than":
                module = self.modules["less_than"]
                bool = module(self.outputs[inputs[0]], self.outputs[inputs[1]])
                # self.outputs[inputs[0]] = None # free up memory
                # self.outputs[inputs[1]] = None # free up memory
                self.outputs[i] = "yes" if bool else "no"
            # count
            elif function == "count":
                module = self.modules["count"]
                mask = self.outputs[inputs[0]]
                if self.use_clean_count_masks:
                    mask = self.make_clean_mask(mask, self.count_clean_mask_threshold)
                log_probs = module({"image": image, "mask": mask})
                n = log_probs.argmax(dim=1).item()
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = n
            # exist
            elif function == "exist":
                module = self.modules["exist"]
                mask = self.outputs[inputs[0]]
                if self.use_clean_count_masks:
                    mask = self.make_clean_mask(mask, self.count_clean_mask_threshold)
                bool, _ = module({"image": image, "mask": mask})
                # self.outputs[inputs[0]] = None # free up memory
                self.outputs[i] = "yes" if bool else "no"
            # unique
            elif function == "unique":
                module = self.modules["unique"]
                mask = self.outputs[inputs[0]]
                orig_mask = mask
                if self.use_clean_count_masks:
                    mask = self.make_clean_mask(mask, self.count_clean_mask_threshold)
                bool, _ = module({"image": image, "mask": mask})
                if self.use_clean_count_masks:
                    self.outputs[i] = orig_mask
                else:
                    self.outputs[i] = mask
        return True

    def forward(self, image, program, answer, idx):
        # initialize the outputs
        self.outputs = [None] * len(program)
        # run the program
        execution_finished = self.run_program(program, image)
        # check if program was able to run all the way through
        if execution_finished:
            # check answer
            final_answer = str(self.outputs[-1])
            logging.info(f"The computed answer is {final_answer}")
            logging.info(f"The ground truth answer is {answer}")
            correct = final_answer == answer
            # export trace if requested and if wrong
            if self.export_trace_pkl is not None and not correct:
                self.export_outputs(image, program, answer, idx)
            # export answer if requested
            if self.export_answers is not None:
                with open(self.export_answers, "a") as f:
                    f.write(final_answer + "\n")
            return correct
        else:
            return False

    def export_outputs(self, image, program, answer, idx):
        export_name, export_ext = os.path.splitext(self.export_trace_pkl)
        export_path = f"{export_name}_{idx}{export_ext}"
        os.makedirs(os.path.dirname(export_path), exist_ok=True)
        with open(export_path, "wb") as f:
            pickle.dump(
                {
                    "image": image.cpu().numpy(),
                    "program": program,
                    "answer": answer,
                    "outputs": self.outputs,
                },
                f,
            )

    def set_mode(self, mode):
        if mode == "train":
            for module in self.modules.values():
                module.train()
        elif mode == "eval":
            for module in self.modules.values():
                module.eval()
