#!/usr/bin/env python

from nmn.datasets.experiment import Experiment
from nmn.datasets.load.utils.transforms import GloveToTensor
from nmn.modules.executors.compare_count import Exists, Unique  # more than one
from nmn.modules.executors.filter import Filter
from nmn.modules.executors.query import Query
from nmn.modules.executors.relate import Relate
from nmn.modules.executors.same import Same
from nmn.modules.load_modules.get_module import get_module
from nmn.modules.subsymbolic.mask_operations import MaskIntersection, MaskUnion
from nmn.modules.symbolic.compare import Equal, GreaterThan, LessThan
from nmn.modules.utils.module_type import ModuleType


def get_clevr_evaluation_modules(checkpoints, config, device):
    # Initialize the modules with the weights from the checkpoint
    # The 'find', 'query' and 'relate' modules are grouped
    # according to the category on which they operate
    # For example, in the category specificity, the 'find'
    # dictionary will contain 'color', 'size', 'shape', etc
    # while in the concept specificity, this will be 'blue',
    # 'brown', 'cylinder', etc.
    pretrained_modules = {
        ModuleType.FIND: {},
        ModuleType.QUERY: {},
        ModuleType.RELATE: {},
    }
    for checkpoint in checkpoints:
        # load module
        module_info = checkpoint["opt"].module_info
        module = get_module(Experiment.CLEVR, device, **module_info)
        state = checkpoint["state"]
        module.load_state_dict(state)

        # put the module in eval mode
        module.eval()
        for param in module.parameters():
            param.requires_grad = False

        # group the modules based on their name
        name = checkpoint["opt"].module_info.name

        if ModuleType.FIND in name:
            key = get_module_key(checkpoint["opt"])
            pretrained_modules[ModuleType.FIND][key] = module
        elif ModuleType.QUERY in name:
            key = get_module_key(checkpoint["opt"])
            pretrained_modules[ModuleType.QUERY][key] = module
        elif ModuleType.RELATE in name:
            if "_" in name:
                # TODO fix: implicit knowledge of name convention
                category = name.split("_")[1]
                pretrained_modules[ModuleType.RELATE][category] = module
            else:
                pretrained_modules[ModuleType.RELATE] = module
        else:
            pretrained_modules[name] = module
    # Create the modules necessary for the programs
    # Some modules rely on others, e.g. Filter uses
    # Find modules, Exist uses the Count module, etc.
    modules = {
        ModuleType.CONTEXT: pretrained_modules[ModuleType.CONTEXT],
        ModuleType.COUNT: pretrained_modules[ModuleType.COUNT],
        "filter": Filter(
            pretrained_modules[ModuleType.FIND], GloveToTensor(config, device)
        ),
        ModuleType.QUERY: Query(pretrained_modules[ModuleType.QUERY]),
        ModuleType.RELATE: Relate(
            pretrained_modules[ModuleType.RELATE], GloveToTensor(config, device)
        ),
        "exist": Exists(pretrained_modules[ModuleType.COUNT]),
        "unique": Unique(pretrained_modules[ModuleType.COUNT]),
        "same": Same(
            pretrained_modules[ModuleType.FIND],
            Query(pretrained_modules[ModuleType.QUERY]),
            GloveToTensor(config, device),
            config,
        ),
        "union": MaskUnion(),
        "intersect": MaskIntersection(),
        "equal": Equal(),
        "less_than": LessThan(),
        "greater_than": GreaterThan(),
    }
    return modules


def get_module_key(opt):
    # TODO fix: implicit knowledge of name convention
    name = opt.module_info.name
    specificity = opt.module_info.specificity
    if specificity == "category":
        return name.split("_")[1]
    elif specificity == "concept":
        return name.split("_")[2]
    else:
        raise ValueError(f"Unexpected specificity {specificity}.")
