#!/usr/bin/env python

import torch
import torch.nn as nn

from nmn.modules.subsymbolic.cnn import ReverseDown


class JeromeNet(nn.Module):
    def __init__(self, **kwargs):
        super(JeromeNet, self).__init__()

        channel_complexity = kwargs.get("channel_complexity")
        n_classes = kwargs.get("n_classes")  # depends on the problem
        fc_width = kwargs.get("fc_width")  # width of the fully connected layers
        depth = kwargs.get("depth")  # amount of down layers
        img_width = kwargs.get("img_width")  # width of the original image
        img_height = kwargs.get("img_height")  # height of the original image

        self.img_conv = ReverseDown(3, 3)  # 3 image channels
        self.att_conv = ReverseDown(1, 3)  # 1 attention channel

        layers = [ReverseDown(3, channel_complexity)]
        # requires depth >= 2
        for i in range(0, depth - 2):
            layers.append(
                ReverseDown(
                    channel_complexity * pow(2, i), channel_complexity * pow(2, i + 1)
                )
            )
        self.down = nn.Sequential(*layers)

        # in channels = A * B * C where
        #    A = width of the final layers (in self.down)
        #    B = original height of the image divided by two with every MaxPool operation in ReverseDown
        #    C = original width of the image divided by two with every MaxPool operation in ReverseDown
        in_channels = (
            (channel_complexity * pow(2, depth - 2))
            * (img_height // (2 ** depth))
            * (img_width // (2 ** depth))
        )

        self.classifier = nn.Sequential(
            nn.Linear(in_channels, fc_width),
            nn.ReLU(inplace=True),
            nn.Linear(fc_width, fc_width),
            nn.ReLU(inplace=True),
            nn.Linear(fc_width, n_classes),
        )

    def forward(self, src):
        img = self.img_conv(src["image"])
        att = self.att_conv(src["mask"])
        x = torch.mul(img, att)
        x = self.down(x)
        x = x.view(x.shape[0], -1)
        return self.classifier(x)
