#!/usr/bin/env python

""" Parts of the U-Net model """

import torch
import torch.nn as nn
import torch.nn.functional as F


class SingleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(SingleConv, self).__init__(*args, **kwargs)
        self.single_conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.single_conv(x)


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(DoubleConv, self).__init__(*args, **kwargs)
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class SingleDown(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(SingleDown, self).__init__(*args, **kwargs)
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2), SingleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class ReverseDown(nn.Module):
    """Single conv then downscaling with maxpool"""

    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(ReverseDown, self).__init__(*args, **kwargs)
        self.maxpool_conv = nn.Sequential(
            SingleConv(in_channels, out_channels), nn.MaxPool2d(2)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(Down, self).__init__(*args, **kwargs)
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2), DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, *args, bilinear=True, **kwargs):
        super(Up, self).__init__(*args, **kwargs)

        # if bilinear, use the normal convolutions
        # to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(
                in_channels // 2, in_channels // 2, kernel_size=2, stride=2
            )

        self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = torch.tensor([x2.size()[2] - x1.size()[2]])
        diffX = torch.tensor([x2.size()[3] - x1.size()[3]])

        halfX = torch.div(diffX, 2, rounding_mode="trunc")
        halfY = torch.div(diffY, 2, rounding_mode="trunc")
        x1 = F.pad(x1, [halfX, diffX - halfX, halfY, diffY - halfY])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super(OutConv, self).__init__(*args, **kwargs)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)
