#!/usr/bin/env python

import torch
import torch.nn as nn
import torch.nn.functional as F

from nmn.modules.subsymbolic.cnn import DoubleConv, OutConv, SingleDown, Up


class PreEncoderOneInput(nn.Module):
    """Pre-encoder to process an image."""

    def __init__(self, **kwargs):
        super().__init__()
        channel_complexity = kwargs.get(
            "channel_complexity"
        )  # historically only find_category used 16, rest
        self.img_conv = DoubleConv(3, channel_complexity)  # 3 image channels

    def forward(self, src):
        return self.img_conv(src["image"])


class PreEncoderTwoInputs(nn.Module):
    """Pre-encoder to process an image and a mask."""

    def __init__(self, **kwargs):
        super().__init__()

        channel_complexity = kwargs.get(
            "channel_complexity"
        )  # historically only find_category used 16, rest
        self.img_conv = DoubleConv(3, channel_complexity)  # 3 image channels
        self.att_conv = DoubleConv(1, channel_complexity)  # 1 attention channel

    def forward(self, src):
        img = self.img_conv(src["image"])
        att = self.att_conv(src["mask"])
        x = torch.mul(img, att)
        return x


class PostDecoder(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

        channel_complexity = kwargs.get(
            "channel_complexity"
        )  # historically only find_category used 16, rest
        self.outc = OutConv(channel_complexity, 1)  # 1 attention channel

    def forward(self, x):
        return self.outc(x)


class TextEncoder(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        x_txt_size = kwargs.get("x_txt_size")
        channel_complexity = kwargs.get(
            "channel_complexity"
        )  # historically only find_category used 16, rest
        depth = kwargs.get("depth")  # amount of up/down layers
        img_width = kwargs.get("img_width")  # width of the original image
        img_height = kwargs.get("img_height")  # height of the original image

        features = pow(2, depth - 1)
        h, w = img_height // (2 ** depth), img_width // (2 ** depth)
        self.lin = nn.Linear(x_txt_size, channel_complexity * features * h * w)

    def forward(self, text, last_down):
        x_txt = self.lin(text)
        x_txt = F.relu(x_txt)
        b, c, h, w = last_down.shape
        x_txt = x_txt.view(b, c, h, w)
        return last_down * x_txt


class Encoder(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

        channel_complexity = kwargs.get(
            "channel_complexity"
        )  # historically only find_category used 16, rest 4
        depth = kwargs.get("depth")  # amount of up/down layers

        layers = []
        for i in range(0, depth):
            if i == depth - 1:
                layers.append(SingleDown(channel_complexity * pow(2, i), channel_complexity * pow(2, i)))
            else:
                layers.append(SingleDown(channel_complexity * pow(2, i), channel_complexity * pow(2, i + 1)))
        self.down = nn.Sequential(*layers)

    def forward(self, x_vis):
        x_down = [x_vis]
        for layer in self.down:
            out = layer(x_down[-1])
            x_down.append(out)

        return x_down


class Decoder(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

        channel_complexity = kwargs.get("channel_complexity")  # only find_category uses 16, rest 4
        depth = kwargs.get("depth")  # amount of down layers

        layers = []
        for i in range(depth, 0, -1):
            if i == 1:
                layers.append(Up(channel_complexity * pow(2, i), channel_complexity))
            else:
                layers.append(Up(channel_complexity * pow(2, i), channel_complexity * pow(2, i - 2)))
        self.up = nn.Sequential(*layers)

    def forward(self, x, skip_connections):
        for layer in self.up:
            x = layer(x, skip_connections.pop())
        return x


class UNet(nn.Module):
    def __init__(self, **kwargs):
        super(UNet, self).__init__()

        self.pre_encoder = self.get_pre_encoder(**kwargs)
        self.specificity = kwargs.get("specificity")
        if self.specificity == "category":
            self.text_key = kwargs.get("category")
            self.text_encoder = TextEncoder(**kwargs)
        self.encoder = Encoder(**kwargs)
        self.decoder = Decoder(**kwargs)
        self.post_decoder = PostDecoder(**kwargs)

    def get_pre_encoder(self, **kwargs):
        input_type = kwargs.get("input_type")
        if input_type == "image":
            self.pre_encoder = PreEncoderOneInput(**kwargs)
        elif input_type == "image+mask":
            self.pre_encoder = PreEncoderTwoInputs(**kwargs)
        else:
            raise ValueError(f"Unexpected input type: {input_type}.")
        return self.pre_encoder

    def forward(self, src):
        x = self.pre_encoder(src)
        x_down = self.encoder(x)
        last_down = x_down.pop()
        x = (
            self.text_encoder(src[self.text_key], last_down)
            if self.specificity == "category"
            else last_down
        )
        x = self.decoder(x, x_down)
        logits = self.post_decoder(x)
        attn = torch.sigmoid(logits)
        return attn
