#!/usr/bin/env python

import json

import numpy as np
import torch
from flask import session
from num2words import num2words

from nmn.datasets.experiment import Experiment
from nmn.server.exceptions import MissingCaseException, MissingKeyException
from nmn.server.utils import compute_iou, get_attention, get_image, same_attention, store_attention


# Base Primitive class
class Primitive(object):
    """Base class for all primitives"""

    def __init__(self, request_data, args):
        """Initialize the primitive"""
        slots = request_data
        # empty arrays are coerced to None
        for key, value in slots.items():
            if value == []:
                slots[key] = None
        # Store the slots and the args
        self.slots = slots
        self.args = args

    def evaluate(self):
        pass

    def maybe_raise_missing_slot_error(self, slot_names):
        """Check if all slot names are present"""
        for slot_name in slot_names:
            if slot_name not in self.slots:
                msg = f"Missing slot '{slot_name}' in primitive '{type(self).__name__,}'"
                raise MissingKeyException(msg)
        return None

    def check_case(self, unbound=[], bound=[]):
        """Check if an IRL case holds"""
        unbound_check = all([self.slots[slot] is None for slot in unbound])
        bound_check = all([self.slots[slot] is not None for slot in bound])
        return unbound_check and bound_check

    def raise_missing_case_error(self):
        """Raise an error for a missing case"""
        bound_slots = [slot for slot in self.slots if self.slots[slot] is not None]
        unbound_slots = [slot for slot in self.slots if self.slots[slot] is None]
        case = f"{bound_slots} => {unbound_slots}"
        msg = f"Missing case for primitive {type(self).__name__}: {case}"
        raise MissingCaseException(msg)


# Context
class Context(Primitive):
    def __init__(self, request_data, args, modules):
        super(Context, self).__init__(request_data, args)
        # get the module
        self.module = modules["context"]
        # check the slots
        self.maybe_raise_missing_slot_error(["context"])

    def evaluate(self):
        """
        The get_context module expects a single slot 'context'.
        (=> context)
        If this slot is unbound, execute the context module and bind
        the resulting attention to 'context'.
        (context =>)
        If this slot is bound, it must be some attention. Execute the
        context module and compare the result to the stored attention.
        """
        # case 1
        if self.check_case(unbound=["context"]):
            attn = self.module({"image": get_image(session)})
            attn_id = store_attention(attn, session)
            return (
                json.dumps({"bindings": [[{"variable": "context", "score": 1.0, "value": attn_id}]]}),
                200,
            )
        # case 2
        elif self.check_case(bound=["context"]):
            stored_attention = get_attention(self.slots["context"], session)
            if self.args.experiment == Experiment.MNIST_DIALOG:
                attn = get_image(session)
            else:
                attn = self.module({"image": get_image(session)})
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Find in Context
class FindInContext(Primitive):
    def __init__(self, request_data, args, modules):
        super(FindInContext, self).__init__(request_data, args)
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "history", "target_attn"])

    def evaluate(self):
        # case 1
        if self.check_case(unbound=["target_attn"]):
            # todo: add multiple history objects
            if (len(self.slots["history"])) == 1:
                attn_id = self.slots["history"][0]
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        else:
            self.raise_missing_case_error()


# Set Difference
class SetDifference(Primitive):
    def __init__(self, request_data, args, modules):
        super(SetDifference, self).__init__(request_data, args)
        # get the module
        self.module = modules["set_diff"]
        # check slots
        self.maybe_raise_missing_slot_error(["source_attn", "history", "target_attn"])

    def evaluate(self):
        """
        Set_diff takes 3 slots: 'source_attn_1', 'source_attn_2'
        and 'target_attn'.
        (source_attn_1 source_attn_2 => target_attn)
        Given source attentions, compute the set-diff as target attn
        (source_attn_1 source_attn_2 target_attn =>)
        Compute the set-diff of the source attentions and compare
        against target attn
        """
        # case 1
        if self.check_case(bound=["source_attn", "history"], unbound=["target_attn"]):
            combined_attn = 0
            for el in self.slots["history"]:
                att = get_attention(el, session)
                combined_attn = torch.add(combined_attn, att, alpha=1)

            source_attn = get_attention(self.slots["source_attn"], session)
            attn = self.module(source_attn, combined_attn)
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "history", "target_attn"]):
            source_attn = get_attention(self.slots["source_attn"], session)
            history = get_attention(self.slots["history"], session)
            attn = self.module(source_attn, history)
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": consistent}), 200
        else:
            self.raise_missing_case_error()


# Count
class Count(Primitive):
    def __init__(self, request_data, args, modules):
        super(Count, self).__init__(request_data, args)
        # get the module
        self.module = modules["count"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "target_num"])

    def evaluate(self):
        """
        The count module gets two slots: 'source_attn' and 'target_num'.
        (source_attn => target_num)
        Run the count module on the attn and return a new binding
        for the target (= an integer)
        (source_attn target_num =>)
        Run the count module on the attn and compare the result
        against target
        """
        # case 1
        if self.check_case(bound=["source_attn"], unbound=["target_num"]):
            log_probs = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                }
            )
            max_log_prob, n = torch.max(log_probs, 1)
            max_prob = np.exp(max_log_prob.item())
            n = n.item()
            return (
                json.dumps({"bindings": [[{"variable": "target_num", "score": max_prob, "value": n}]]}),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_num"]):
            log_probs = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                }
            )
            n = log_probs.argmax(dim=1).item()
            consistent = n == self.slots["target_num"]
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Exist
class Exist(Primitive):
    def __init__(self, request_data, args, modules):
        super(Exist, self).__init__(request_data, args)
        # get the module
        self.module = modules["exist"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "target_bool"])

    def evaluate(self):
        """
        The exist module takes 2 slots: 'source_attn' and 'target_bool'.
        (source_attn => target_bool)
        Run the exist module on the source attn and return a new binding
        for target bool
        (source_attn target_bool =>)
        Run the exist module on the source attn and compare against
        target bool
        """
        # case 1
        if self.check_case(bound=["source_attn"], unbound=["target_bool"]):
            bool, prob = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                }
            )
            target = "yes" if bool else "no"
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_bool",
                                    "score": prob,
                                    "value": target,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_bool"]):
            bool, _ = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                }
            )
            target = "yes" if bool else "no"
            consistent = target == self.slots["target_bool"]
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Filter
class Filter(Primitive):
    def __init__(self, request_data, args, modules):
        super(Filter, self).__init__(request_data, args)
        # get the module
        self.module = modules["filter"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "target_attn", "concept"])

    def evaluate(self):
        """
        The filter primitive takes 3 slots: 'source_attn',
        'target_attn' and 'concept'.
        (source_attn concept => target_attn)
        Run the filter module and bind the target attn
        (source_attn target_attn => concept)
        Try the filter module with all concepts and compare
        against the target attn
        (source_attn => target_attn concept)
        Run the filter module with all concept and create
        many bindings
        (source_attn target_attn concept =>)
        Run the filter module with the concept and compare
        against the target attn
        """
        # case 1
        if self.check_case(bound=["source_attn", "concept"], unbound=["target_attn"]):
            # the module needs to have both the 'category' (e.g. color)
            # and the 'concept' (e.g. blue). IRL always uses the latter,
            # so we have to retrieve the former
            filter_category = None
            for category, concepts in self.args.config.categories.items():
                if str(self.slots["concept"]) in concepts:
                    filter_category = category
                    break
            # !!! exceptions
            if self.slots["concept"] == "thing":
                # filter_thing in CLEVR is just returning the same attention
                attn = get_attention(self.slots["source_attn"], session)
            elif self.slots["concept"] == "number":
                # filter_number in MNIST is just returning the same attention
                attn = get_attention(self.slots["source_attn"], session)
            else:
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "category": filter_category,
                        "concept": str(self.slots["concept"]),
                    }
                )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_attn"], unbound=["concept"]):
            possible_values = []
            stored_attention = get_attention(self.slots["target_attn"], session)
            for category, concepts in self.args.config.categories.items():
                for concept in concepts:
                    attn = self.module(
                        {
                            "image": get_image(session),
                            "mask": get_attention(self.slots["source_attn"], session),
                            "category": category,
                            "concept": concept,
                        }
                    )
                    iou = compute_iou(stored_attention, attn)
                    if iou > self.args.iou_threshold:
                        possible_values.append((concept, iou))
            # here we choose to return a single value
            # other possibilty would be to return all concepts
            # that have iou > threshold, and use the iou
            # as score in IRL
            best_value, best_iou = max(possible_values, key=lambda p: p[1])
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "concept",
                                    "score": best_iou,
                                    "value": best_value,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 3
        elif self.check_case(bound=["source_attn"], unbound=["target_attn", "concept"]):
            bindings = []
            # run the find module for all categories and concepts
            # and return all bindings
            for category, concepts in self.args.config.categories.items():
                for concept in concepts:
                    attn = self.module(
                        {
                            "image": get_image(session),
                            "mask": get_attention(self.slots["source_attn"], session),
                            "category": category,
                            "concept": concept,
                        }
                    )
                    attn_id = store_attention(attn, session)
                    bindings.append(
                        [
                            {"variable": "target_attn", "score": 1.0, "value": attn_id},
                            {"variable": "concept", "score": 1.0, "value": concept},
                        ]
                    )
            return json.dumps({"bindings": bindings}), 200
        # case 4
        elif self.check_case(bound=["source_attn", "target_attn", "concept"]):
            for category, concepts in self.args.config.categories.items():
                if self.slots["concept"] in concepts:
                    filter_category = category
                    break

            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "category": filter_category,
                    "concept": self.slots["concept"],
                }
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Intersect
class Intersect(Primitive):
    def __init__(self, request_data, args, modules):
        super(Intersect, self).__init__(request_data, args)
        # get the module
        self.module = modules["intersect"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn_1", "source_attn_2", "target_attn"])

    def evaluate(self):
        """
        Intersect takes 3 slots: 'source_attn_1', 'source_attn_2'
        and 'target_attn'.
        (source_attn_1 source_attn_2 => target_attn)
        Given source attentions, compute the intersection as target attn
        (source_attn_1 source_attn_2 target_attn =>)
        Compute the intersection of the source attentions and compare
        against target attn
        """
        # case 1
        if self.check_case(bound=["source_attn_1", "source_attn_2"], unbound=["target_attn"]):
            attn = self.module(
                get_attention(self.slots["source_attn_1"], session),
                get_attention(self.slots["source_attn_2"], session),
            )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        elif self.check_case(bound=["source_attn_1", "source_attn_2", "target_attn"]):
            attn = self.module(
                get_attention(self.slots["source_attn_1"], session),
                get_attention(self.slots["source_attn_2"], session),
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Query
class Query(Primitive):
    def __init__(self, request_data, args, modules):
        super(Query, self).__init__(request_data, args)
        # get the module
        self.module = modules["query"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "attribute", "target_concept"])

    def evaluate(self):
        """
        Query takes 3 slots: 'source_attn', 'attribute' and
        'target_concept'.
        (source_attn attribute => target_concept)
        (source_attn target_concept => attribute)
        (source_attn => attribute target_concept)
        (source_attn attribute target_concept =>)
        """
        # case 1
        if self.check_case(bound=["source_attn", "attribute"], unbound=["target_concept"]):
            options = self.args.config.categories[self.slots["attribute"]]
            index, prob = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "category": self.slots["attribute"],
                    "options": options,
                }
            )
            label = self.args.config.categories[self.slots["attribute"]][index]
            if self.args.experiment == Experiment.MNIST_DIALOG:
                label = num2words(label)
                if self.slots["attribute"] == "bgcolor":
                    label = label + "-bg"
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_concept",
                                    "score": prob,
                                    "value": label,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_concept"], unbound=["attribute"]):
            bindings = []
            for category, concepts in self.args.config.categories.items():
                index, prob = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "category": category,
                        "options": concepts,
                    }
                )
                label = self.args.config.categories[category][index]
                if label == self.slots["target_concept"]:
                    bindings.append([{"variable": "attribute", "score": prob, "value": category}])
            return json.dumps({"bindings": bindings}), 200
        # case 3
        elif self.check_case(bound=["source_attn"], unbound=["attribute", "target_concept"]):
            bindings = []
            for category, concepts in self.args.config.categories.items():
                index, prob = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "category": category,
                        "options": concepts,
                    }
                )
                label = self.args.config.categories[category][index]
                bindings.append(
                    [
                        {"variable": "attribute", "score": 1.0, "value": category},
                        {"variable": "target_concept", "score": prob, "value": label},
                    ]
                )
            return json.dumps({"bindings": bindings}), 200
        # case 4
        elif self.check_case(bound=["source_attn", "attribute", "target_concept"]):
            options = self.args.config.categories[self.slots["attribute"]]
            index, _ = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "category": self.slots["attribute"],
                    "options": options,
                }
            )
            label = self.args.config.categories[self.slots["attribute"]][index]
            consistent = label == self.slots["target_concept"]
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Relate
class Relate(Primitive):
    def __init__(self, request_data, args, modules):
        super(Relate, self).__init__(request_data, args)
        # get the module
        self.module = modules["relate"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "spatial_relation", "target_attn"])

    def evaluate(self):
        """
        Relate takes 3 slots: 'source_attn', 'spatial_relation'
        and 'target_attn'.
        (source_attn spatial_relation => target_attn)
        (source_attn target_attn => spatial_relation)
        (source_attn => spatial_relation target_attn)
        (source_attn spatial_relation target_attn =>)
        """
        # case 1
        if self.check_case(bound=["source_attn", "spatial_relation"], unbound=["target_attn"]):
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "relationship": self.slots["spatial_relation"],
                }
            )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_attn"], unbound=["spatial_relation"]):
            possible_relations = []
            stored_attention = get_attention(self.slots["target_attn"], session)
            for relation in self.args.config.relate:
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "relationship": relation,
                    }
                )
                iou = compute_iou(stored_attention, attn)
                if iou > self.args.iou_threshold:
                    possible_relations.append((relation, iou))
            best_relation, best_iou = max(possible_relations, key=lambda p: p[1])
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "spatial_relation",
                                    "score": best_iou,
                                    "value": best_relation,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 3
        elif self.check_case(bound=["source_attn"], unbound=["spatial_relation", "target_attn"]):
            bindings = []
            for relation in self.args.config.relate:
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "relationship": relation,
                    }
                )
                attn_id = store_attention(attn, session)
                bindings.append(
                    [
                        {
                            "variable": "spatial_relation",
                            "score": 1.0,
                            "value": relation,
                        },
                        {"variable": "target_attn", "score": 1.0, "value": attn_id},
                    ]
                )
            return json.dumps({"bindings": bindings}), 200
        # case 4
        elif self.check_case(bound=["source_attn", "spatial_relation", "target_attn"]):
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "relationship": self.slots["spatial_relation"],
                }
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Same
class Same(Primitive):
    def __init__(self, request_data, args, modules):
        super(Same, self).__init__(request_data, args)
        # get the module
        self.module = modules["same"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "attribute", "target_attn"])

    def evaluate(self):
        """
        Same takes 3 slots: 'source_attn', 'attribute'
        and 'target_attn'
        (source_attn attribute => target_attn)
        (source_attn target_attn => attribute)
        (source_attn => attribute target_attn)
        (source_attn attribute target_attn =>)
        """
        # case 1
        if self.check_case(bound=["source_attn", "attribute"], unbound=["target_attn"]):
            options = self.args.config.categories[self.slots["attribute"]]
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "category": self.slots["attribute"],
                    "options": options,
                }
            )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_attn"], unbound=["attribute"]):
            possible_values = []
            stored_attention = get_attention(self.slots["target_attn"], session)
            for category, options in self.args.config.categories.items():
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "category": category,
                        "options": options,
                    }
                )
                iou = compute_iou(stored_attention, attn)
                if iou > self.argsÒ.iou_threshold:
                    possible_values.append((category, iou))
            best_value, best_iou = max(possible_values, key=lambda p: p[1])
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "attribute",
                                    "score": best_iou,
                                    "value": best_value,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 3
        elif self.check_case(bound=["source_attn"], unbound=["attribute", "target_attn"]):
            bindings = []
            for category, options in self.args.config.categories.items():
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "category": category,
                        "options": options,
                    }
                )
                attn_id = store_attention(attn, session)
                bindings.append(
                    [
                        {"variable": "attribute", "score": 1.0, "value": category},
                        {"variable": "target_attn", "score": 1.0, "value": attn_id},
                    ]
                )
            return json.dumps({"bindings": bindings}), 200
        # case 4
        elif self.check_case(bound=["source_attn", "attribute", "target_attn"]):
            options = self.args.config.categories[self.slots["attribute"]]
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "category": self.slots["attribute"],
                    "options": options,
                }
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Union
class Union(Primitive):
    def __init__(self, request_data, args, modules):
        super(Union, self).__init__(request_data, args)
        # get the module
        self.module = modules["union"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn_1", "source_attn_2", "target_attn"])

    def evaluate(self):
        """
        Union takes 3 slots: 'source_attn_1', 'source_attn_2'
        and 'target_attn'.
        (source_attn_1 source_attn_2 => target_attn)
        Given source attentions, compute the union as target attn
        (source_attn_1 source_attn_2 target_attn =>)
        This requires to compare attentions
        """
        # case 1
        if self.check_case(bound=["source_attn_1", "source_attn_2"], unbound=["target_attn"]):
            attn = self.module(
                get_attention(self.slots["source_attn_1"], session),
                get_attention(self.slots["source_attn_2"], session),
            )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        elif self.check_case(bound=["source_attn_1", "source_attn_2", "target_attn"]):
            attn = self.module(
                get_attention(self.slots["source_attn_1"], session),
                get_attention(self.slots["source_attn_2"], session),
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()


# Unique
class Unique(Primitive):
    def __init__(self, request_data, args, modules):
        super(Unique, self).__init__(request_data, args)
        # get the module
        self.module = modules["unique"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "target_attn"])

    def evaluate(self):
        """
        Unique has 2 slots: 'source_attn' and 'target_attn'
        (source_attn => target_attn)
        (source_attn target_attn =>)
        """
        # case 1
        if self.check_case(bound=["source_attn"], unbound=["target_attn"]):
            att = get_attention(self.slots["source_attn"], session)
            bool, prob = self.module({"image": get_image(session), "mask": att})
            if bool:
                # if unique, return the source attn as target attn
                return (
                    json.dumps(
                        {
                            "bindings": [
                                [
                                    {
                                        "variable": "target_attn",
                                        "score": prob,
                                        "value": self.slots["source_attn"],
                                    }
                                ]
                            ]
                        }
                    ),
                    200,
                )
            else:
                # return no bindings, forcing IRL to stop
                return json.dumps({"bindings": []}), 200
        # case 2
        elif self.check_case(bound=["source_attn", "target_attn"]):
            attn = get_attention(self.slots["source_attn"], session)
            bool, _ = self.module({"image": get_image(session), "mask": attn})
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn, self.args.iou_threshold)
            return json.dumps({"consistent": bool and consistent}), 200
        else:
            self.raise_missing_case_error()


# Immediate Relate
class ImmediateRelate(Primitive):
    def __init__(self, request_data, args, modules):
        super(ImmediateRelate, self).__init__(request_data, args)
        # get the module
        self.module = modules["immediate_relate"]
        # check the slots
        self.maybe_raise_missing_slot_error(["source_attn", "spatial_relation", "target_attn"])

    def evaluate(self):
        """
        Relate takes 3 slots: 'source_attn', 'spatial_relation'
        and 'target_attn'.
        (source_attn spatial_relation => target_attn)
        (source_attn target_attn => spatial_relation)
        (source_attn => spatial_relation target_attn)
        (source_attn spatial_relation target_attn =>)
        """
        # case 1
        if self.check_case(bound=["source_attn", "spatial_relation"], unbound=["target_attn"]):
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "relationship": self.slots["spatial_relation"],
                }
            )
            attn_id = store_attention(attn, session)
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": 1.0,
                                    "value": attn_id,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 2
        elif self.check_case(bound=["source_attn", "target_attn"], unbound=["spatial_relation"]):
            possible_relations = []
            stored_attention = get_attention(self.slots["target_attn"], session)
            for relation in self.args.config.relate:
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "relationship": relation,
                    }
                )
                iou = compute_iou(stored_attention, attn)
                if iou > self.args.iou_threshold:
                    possible_relations.append((relation, iou))
            best_relation, best_iou = max(possible_relations, key=lambda p: p[1])
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "spatial_relation",
                                    "score": best_iou,
                                    "value": best_relation,
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        # case 3
        elif self.check_case(bound=["source_attn"], unbound=["spatial_relation", "target_attn"]):
            bindings = []
            for relation in self.args.config.relate:
                attn = self.module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(self.slots["source_attn"], session),
                        "relationship": relation,
                    }
                )
                attn_id = store_attention(attn, session)
                bindings.append(
                    [
                        {
                            "variable": "spatial_relation",
                            "score": 1.0,
                            "value": relation,
                        },
                        {"variable": "target_attn", "score": 1.0, "value": attn_id},
                    ]
                )
            return json.dumps({"bindings": bindings}), 200
        # case 4
        elif self.check_case(bound=["source_attn", "spatial_relation", "target_attn"]):
            attn = self.module(
                {
                    "image": get_image(session),
                    "mask": get_attention(self.slots["source_attn"], session),
                    "relationship": self.slots["spatial_relation"],
                }
            )
            stored_attention = get_attention(self.slots["target_attn"], session)
            consistent = same_attention(stored_attention, attn)
            return json.dumps({"consistent": str(consistent)}), 200
        else:
            self.raise_missing_case_error()
