#!/usr/bin/env python

import logging

import numpy as np
from sklearn.metrics import jaccard_score

import nmn.server.globals as globals
from nmn.modules.load_modules.get_modules import get_evaluation_modules


def load_modules(args):
    """
    Load all modules according to args
    """
    kwargs = {
        "specificity": args.specificity,
        "experiment": args.experiment,
        "config": args.config,
        "device": "cpu",
    }
    globals.modules = get_evaluation_modules(**kwargs)
    logging.info("Loaded the following modules:")
    logging.info(f"{globals.modules.keys()}")


def check_request_data(request_data, keys):
    """
    Helper function that checks if key(s) are present
    in the data and if they are, returns the data.
    If they are not, a MissingKey error is raised!
    """
    for key in keys:
        if key not in request_data:
            msg = "No data provided using the '%s' key" % key
            return msg
    return None


def get_image(session):
    """Retrieve the current image"""
    return session["current_scene"]


def store_attention(attn, session):
    """Store a new attention"""
    globals.attention_counter += 1
    name = "attn_" + str(globals.attention_counter)
    session["attentions"][name] = attn
    return name


def get_attention(attn_id, session):
    """Retrieve an attention by its ID"""
    return session["attentions"][attn_id]


def same_attention(attn_1, attn_2, threshold):
    """
    Check if two attentions are the same
    according to the specified threshold.
    """
    iou = compute_iou(attn_1, attn_2)
    return iou >= threshold


def compute_iou(attn_1, attn_2):
    """Computes the IOU"""
    # Discretize the attentions
    mask_1 = np.where(attn_1 <= 0.5, 0, 1)
    mask_2 = np.where(attn_2 <= 0.5, 0, 1)
    # Flatten to single dim
    mask_1 = mask_1.reshape(-1)
    mask_2 = mask_2.reshape(-1)
    # Compute IOU
    iou = jaccard_score(mask_1, mask_2, average="macro")
    return iou


def clear_all_attentions(session):
    """Clear all attentions stored in the session"""
    if "attentions" in session:
        num_stored_attentions = len(session["attentions"])
        logging.info(f"Clearing {num_stored_attentions} attentions.")
    session["attentions"] = {}
