#!/usr/bin/env python

import json
import logging
import os
import random

import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter


class Checkpointer:
    """The Checkpointer records and logs data produced during the training process."""

    def __init__(self, opt, model, module_kwargs, dataset_size, checkpoint=None):
        # options
        self.print_every = opt.print_every
        self.record_every = opt.record_every
        self.export_mode = opt.export_mode
        self.mode = opt.mode
        self.model = model

        # epochs already processed in past
        self.last_epoch = checkpoint["epoch"][-1] if checkpoint is not None else 0
        self.dataset_size = dataset_size
        self.train_losses = []

        # checkpointing and tb logging
        pt_name = f"{'_'.join([str(el) for el in list(opt.module_info.values())])}_{random.randint(1, 1000000000)}"
        # checkpoint
        opt.checkpoint_path = os.path.join(opt.logdir, "checkpoints", pt_name + ".pt")
        self.checkpoint_path = opt.checkpoint_path
        os.makedirs(os.path.dirname(opt.checkpoint_path), exist_ok=True)
        logging.info(f"Checkpoint path is {opt.checkpoint_path}")
        # tb logging
        opt.tb_path = os.path.join(opt.logdir, "tensorboard", pt_name)
        self.tb_path = opt.tb_path
        self.writer = SummaryWriter(log_dir=opt.tb_path)
        logging.info(f"Tensorboard path is {opt.tb_path}")

        self.checkpoint = self.initialize_checkpoint(opt, module_kwargs, checkpoint)

    def initialize_checkpoint(self, opt, module_kwargs, checkpoint=None):
        return {
            # Store the command line options
            "opt": {
                "experiment": opt.experiment,
                "module_info": dict(module_kwargs),
                "training": {
                    "use_cuda": opt.use_cuda,
                    "num_workers": opt.num_workers,
                    "num_epochs": opt.num_epochs,
                    "batch_size": opt.batch_size,
                    "shuffle_data": opt.shuffle_data,
                    "learning_rate": opt.learning_rate,
                    "weight_decay": opt.weight_decay,
                    "max_samples_train": opt.max_samples_train,
                    "max_samples_val": opt.max_samples_val,
                    "balance_classes": opt.balance_classes,
                    "use_early_stopping": opt.use_early_stopping,
                    "early_stopping_patience": opt.early_stopping_patience,
                    "early_stopping_delta": opt.early_stopping_delta,
                    "early_stopping_verbose": opt.early_stopping_verbose,
                },
                "logging": {
                    "print_every": opt.print_every,
                    "record_every": opt.record_every,
                    "device": opt.device,
                    "checkpoint_path": opt.checkpoint_path,
                    "tb_path": opt.tb_path,
                },
            },
            "state": checkpoint["state"] if checkpoint else None,
            "stats": {
                "epochs": checkpoint["epochs"] if checkpoint else [],
                "train_losses": checkpoint["train_losses"] if checkpoint else [],
                "val_losses": checkpoint["val_losses"] if checkpoint else [],
                "val_accuracies": checkpoint["val_accuracies"] if checkpoint else [],
                "state_epoch": checkpoint["state_epoch"] if checkpoint else None,
                "state_val_loss": checkpoint["state_val_loss"] if checkpoint else None,
            },
        }

    def init(self):
        if self.mode == "train":
            logging.info("===== Start training =====")
        elif self.mode == "resume":
            logging.info("===== Resume training =====")

    def close(self):
        # close tb writer
        self.writer.close()
        logging.info("===== Finished training =====")

    def log_train(self, epoch, t, loss):
        sample_idx = ((epoch + self.last_epoch) * self.dataset_size) + t

        self.train_losses.append(loss)
        if (t % self.print_every) == 0:
            logging.info(f"  | {sample_idx + 1} {loss}")
        if (t % self.record_every) == 0:
            self.writer.add_scalar("train/loss", loss, sample_idx)

    def save_checkpoint(self, checkpoint, path):
        logging.info(f"  | Saving checkpoint to {path}")
        torch.save(checkpoint, path)

        # following code just copies the data to a json in a readable format
        # eases debugging
        json_checkpoint = {k: v for k, v in checkpoint.items()}
        # state and config are not saved
        json_checkpoint["state"] = None
        with open(path + ".json", "w") as f:
            json.dump(json_checkpoint, f)

    def set_checkpoint_state(self, epoch, val_loss):
        self.checkpoint["state"] = self.model.get_state()
        self.checkpoint["stats"]["state_epoch"] = epoch
        self.checkpoint["stats"]["state_val_loss"] = val_loss

    def generate_checkpoint_with_mode(self, current_epoch, val_loss):
        if self.export_mode == "all":
            # We want all checkpoints, so we give each of them
            # a unique name by adding the timestep
            name, ext = os.path.splitext(self.checkpoint_path)
            export_path = f"{name}_epoch{current_epoch+1}{ext}"
            self.set_checkpoint_state(current_epoch, val_loss)
        elif self.export_mode == "last":
            # we only want the last checkpoint, so overwrite it each time
            export_path = self.checkpoint_path
            self.set_checkpoint_state(current_epoch, val_loss)
        elif self.export_mode == "best":
            # we only want the best checkpoint, so check the
            # best_val_acc before exporting
            if (
                self.checkpoint["stats"]["state_val_loss"] is None
                or val_loss < self.checkpoint["stats"]["state_val_loss"]
            ):
                export_path = self.checkpoint_path
                self.set_checkpoint_state(current_epoch, val_loss)
        else:
            raise ValueError(f"Unexpected export_mode: {self.export_mode}")

        return export_path

    def log_validation(self, epoch, val_loss, val_acc):
        current_epoch = epoch + self.last_epoch
        sample_idx = current_epoch * self.dataset_size

        train_mean_loss = np.mean(self.train_losses).item()
        self.train_losses = []  # reset train losses

        logging.info(f"  | training loss is {train_mean_loss}")
        logging.info(f"  | validation loss is {val_loss}")
        logging.info(f"  | validation accuracy is {val_acc}")

        self.writer.add_scalar("val/loss", val_loss, sample_idx)
        self.writer.add_scalar("val/acc", val_acc, sample_idx)

        self.checkpoint["stats"]["epochs"].append(current_epoch)
        self.checkpoint["stats"]["train_losses"].append(train_mean_loss)
        self.checkpoint["stats"]["val_losses"].append(val_loss)
        self.checkpoint["stats"]["val_accuracies"].append(val_acc)

        path = self.generate_checkpoint_with_mode(current_epoch, val_loss)
        self.save_checkpoint(self.checkpoint, path)
