#!/usr/bin/env python

import torch.nn as nn
from torch.optim import Adam


class Model(nn.Module):
    """
    The Model class represents the module that is being trained,
    together with its loss function, evaluation function and
    optimiser. The trainer passes data to the model for training.
    """

    def __init__(
        self,
        module,
        loss_fn,
        evaluation_fn,
        learning_rate=None,
        weight_decay=None,
    ):
        super(Model, self).__init__()
        self.module = module
        self.loss_fn = loss_fn
        self.evaluation_fn = evaluation_fn
        if learning_rate is not None and weight_decay is not None:
            self.optimizer = Adam(
                self.module.parameters(), lr=learning_rate, weight_decay=weight_decay
            )

    def forward(self, source, target, mode="train"):
        prediction = self.module(source)
        loss = self.loss_fn(prediction, target)
        if mode == "train":
            return loss
        elif mode == "val":
            accuracy = self.evaluation_fn(prediction, target)
            return loss, accuracy
        elif mode == "test":
            accuracy = self.evaluation_fn(prediction, target)
            return loss, accuracy, prediction

    def count_parameters(self):
        return sum(p.numel() for p in self.module.parameters() if p.requires_grad)

    def set_mode(self, mode):
        if mode == "train":
            self.module.train()
        elif mode == "eval":
            self.module.eval()

    def get_state(self):
        state = {}
        for k, v in self.module.state_dict().items():
            state[k] = v.clone()
        return state
