#!/usr/bin/env python

import logging

import numpy as np
import torch

from nmn.train.checkpointer import Checkpointer
from nmn.train.early_stopping import EarlyStopping


class Trainer(object):
    """
    The Trainer specifies the training loop. It gets data from the
    train_loader and passes it to the model, which computes the loss.
    Every so often, the validation loss is computed and a checkpoint
    is created.
    """

    def __init__(
        self, opt, model, module_kwargs, train_loader, val_loader, checkpoint=None
    ):
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.model = model
        self.checkpointer = Checkpointer(
            opt, model, module_kwargs, len(train_loader), checkpoint
        )

        # Initialize the early stopping mechanism if requested
        self.early_stopper = (
            EarlyStopping(
                patience=opt.early_stopping_patience,
                delta=opt.early_stopping_delta,
                verbose=opt.early_stopping_verbose,
            )
            if opt.use_early_stopping
            else False
        )

        # if resuming training, retrieve timestep and epoch data
        self.num_epochs = opt.num_epochs
        self.epoch = checkpoint["epoch"][-1] if checkpoint is not None else 0
        self.device = opt.device

    def run(self):
        self.checkpointer.init()
        stop_early = False
        while self.epoch < self.num_epochs and not stop_early:
            self.train()
            val_loss = self.val()
            if self.early_stopper:
                stop_early = self.early_stopper(val_loss)
            self.epoch += 1
        self.checkpointer.close()

    def train(self):
        self.model.set_mode("train")
        logging.info(f" | Epoch {self.epoch + 1}")
        for t, batch in enumerate(self.train_loader, 0):
            # feed sample
            source, target = batch
            target = target.to(self.device)
            self.model.optimizer.zero_grad()
            loss = self.model.forward(source, target)
            loss.backward()
            self.model.optimizer.step()
            # log training
            self.checkpointer.log_train(self.epoch, t, loss.item())

    def val(self):
        logging.info(" | Running validation")
        self.model.set_mode("eval")
        losses = []
        accuracies = []
        for batch in self.val_loader:
            source, target = batch
            target = target.to(self.device)
            with torch.no_grad():
                loss, accuracy = self.model.forward(source, target, mode="val")
            losses.append(loss.item())
            accuracies.append(accuracy)

        avg_val_loss = np.mean(losses).item()
        avg_val_acc = np.mean(accuracies).item()

        self.checkpointer.log_validation(self.epoch, avg_val_loss, avg_val_acc)

        return avg_val_loss
