import json
import logging
import os

import cv2
import pycocotools.mask
from detectron2 import model_zoo
from detectron2.config import get_cfg
from detectron2.data import DatasetCatalog, MetadataCatalog
from detectron2.engine import DefaultPredictor
from detectron2.structures import BoxMode


def change_shape_to_cat_id(shape):
    if shape == "sphere":
        return 0
    elif shape == "cylinder":
        return 1
    elif shape == "cube":
        return 2


def get_clevr_dicts():
    CLEVR_mini_dir = "data/clevr_mini"
    json_file = os.path.join(CLEVR_mini_dir, "CLEVR_mini_coco_anns.json")
    img_dir = os.path.join(CLEVR_mini_dir, "images")
    dataset_dicts = []

    with open(json_file) as f:

        imgs_anns = json.load(f)
        scenes = imgs_anns["scenes"]
        for ind, scene in enumerate(scenes):
            record = {}

            filename = os.path.join(img_dir, scene["image_filename"])
            height, width = cv2.imread(filename).shape[:2]

            record["file_name"] = filename
            record["image_id"] = ind
            record["height"] = height
            record["width"] = width

            objs = []

            for o in scene["objects"]:
                obj = {
                    "bbox": list(pycocotools.mask.toBbox([o["mask"]])[0]),
                    "bbox_mode": BoxMode.XYWH_ABS,
                    "segmentation": o["mask"],
                    "category_id": change_shape_to_cat_id(o["shape"]),
                }
                objs.append(obj)
            record["annotations"] = objs
            dataset_dicts.append(record)

    return dataset_dicts


def load_clevr_metadata():
    remove_clevr_mini = False
    for el in DatasetCatalog:
        if el == "clevr_mini_train":
            remove_clevr_mini = True
    if remove_clevr_mini:
        DatasetCatalog.remove("clevr_mini_train")
    DatasetCatalog.register("clevr_mini_train", get_clevr_dicts())
    MetadataCatalog.get("clevr_mini_train").thing_classes = [
        "sphere",
        "cylinder",
        "cube",
    ]
    clevr_metadata = MetadataCatalog.get("clevr_mini_train")
    return clevr_metadata


def get_detectron_path(args):
    """
    Parse the command line args to obtain a particular
    pretrained model from Detectron's model zoo
    """
    logging.info(f"the model is: {args.model}")
    if args.model == "mask_rcnn":
        return (
            "_".join(
                ["COCO-InstanceSegmentation/mask_rcnn", args.backbone, args.schedule]
            )
            + ".yaml"
        )
    elif args.model == "faster_rcnn":
        return (
            "_".join(["COCO-Detection/faster_rcnn", args.backbone, args.schedule])
            + ".yaml"
        )


def load_predictor(device, **kwargs):
    """Instantiates a Mask R-CNN instance with a set of pretrained weights.

    Args:
        detectron_weights_path (str): path to the .pt file with the weights of the model

    Returns:
        DefaultPredictor: a Detectron2 predictor
    """
    cfg = get_cfg()
    cfg.merge_from_file(model_zoo.get_config_file(kwargs.get("architecture_path")))

    cfg.DATALOADER.NUM_WORKERS = 0
    cfg.MODEL.DEVICE = device
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = kwargs.get("n_classes")
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = kwargs.get("threshold")
    cfg.INPUT.MASK_FORMAT = kwargs.get("mask_format")
    cfg.MODEL.WEIGHTS = kwargs.get("weights")

    predictor = DefaultPredictor(cfg)
    return predictor
