#!/usr/bin/env python

import logging

import torch

from nmn.datasets.load.dataloader import get_evaluation_dataloader
from nmn.datasets.load.experiment_dataset import collate_fn
from nmn.evaluate.evaluate_experiment import ExperimentEvaluator
from nmn.evaluate.program_executor import ProgramExecutor
from nmn.modules.load_modules.get_modules import get_evaluation_modules
from nmn.options.get_options import get_options
from nmn.utils.config import load_experiment_config
from nmn.utils.logger import setup_logger
from scripts.utils.loading import get_clevr_questions, get_images


def main():
    """Script serves to solely evaluate the CLEVR (VQA) experiment."""
    # Parse the command line args
    opt = get_options("eval")

    # Read the config file
    load_experiment_config(opt)

    # log run
    setup_logger(opt)

    # Make the dataloader
    dataloader_kwargs = {
        "experiment": opt.experiment,
        "questions": get_clevr_questions(),
        "images_dir": get_images(opt.experiment, "test"),
        "device": opt.device,
        "max_samples": opt.max_samples,
        "shuffle": opt.shuffle_data,
        "batch_size": 1,
        "num_workers": opt.num_workers,
        "collate_fn": collate_fn,
    }
    data_loader = get_evaluation_dataloader(**dataloader_kwargs)
    logging.info(f"Data loader has {len(data_loader)} samples")

    # Get the correct modules,
    # depending on the specificity
    # that is evaluated
    module_kwargs = {
        "specificity": opt.specificity,
        "experiment": opt.experiment,
        "config": opt.config,
        "device": opt.device,
    }
    modules = get_evaluation_modules(**module_kwargs)
    logging.info("Modules loaded")

    # Create the ProgramExecutor
    program_executor = ProgramExecutor(
        modules,
        opt.specificity,
        opt.config,
        opt.export_trace_pkl,
        opt.export_answers,
        opt.use_clean_count_masks,
        opt.use_clean_query_masks,
        opt.count_clean_mask_threshold,
        opt.query_clean_mask_threshold,
    )
    program_executor.set_mode("eval")
    logging.info("ProgramExecutor created")

    # Create the Evaluator and run!
    evaluator = ExperimentEvaluator(program_executor, data_loader, opt.evaluation_mode)
    logging.info("ExperimentEvaluator created")
    evaluator.evaluate()


if __name__ == "__main__":
    with torch.no_grad():
        main()
