#!/usr/bin/env python

import logging

import torch

from nmn.datasets.load.dataloader import get_dataloader
from nmn.evaluate.evaluate_module import ModuleEvaluator
from nmn.evaluate.evaluation_metrics import get_evaluation_fn
from nmn.modules.load_modules.get_module import get_module
from nmn.options.get_options import get_options
from nmn.train.loss import get_loss_fn
from nmn.train.model import Model
from nmn.utils.config import load_experiment_config
from nmn.utils.dotdict import DotDict
from nmn.utils.logger import setup_logger
from scripts.utils.loading import get_images, get_scenes


def main():
    # Parse command line args
    opt = get_options("test")

    # Read the config file
    load_experiment_config(opt)

    # log run
    setup_logger(opt)

    # Read the checkpoint
    checkpoint = torch.load(opt.checkpoint, map_location=lambda storage, loc: storage)
    pt_opt = DotDict(checkpoint["opt"])
    logging.info(pt_opt)
    logging.info("Checkpoint loaded")

    # Get test dataloaders
    test_loader_kwargs = {
        "device": opt.device,
        # dataloader options
        "num_workers": opt.num_workers,
        "shuffle": opt.shuffle_data,
        "batch_size": opt.batch_size,
        # test options
        "images": get_images(opt.experiment, "test"),
        "scenes": get_scenes(opt.experiment, "test", pt_opt.module_info.name),
        "max_samples": opt.max_samples_test,
        # pretrained modules
        "module_info": pt_opt.module_info,
        # config
        "config": opt.config,
    }
    test_loader = get_dataloader(**test_loader_kwargs)
    logging.info(f"test_loader has {len(test_loader.dataset)} samples")

    # Get the module
    module_kwargs = pt_opt.module_info
    module = get_module(opt.experiment, opt.device, **module_kwargs)
    state = checkpoint["state"]
    module.load_state_dict(state)

    # Get the loss function
    loss_fn = get_loss_fn(pt_opt.module_info.type)

    # Get the evaluation function
    evaluation_fn = get_evaluation_fn(pt_opt.module_info.type)

    # Use all of the above to create the model
    # and initialize it with the weights from the checkpoint
    model = Model(
        module=module,
        loss_fn=loss_fn,
        evaluation_fn=evaluation_fn,
    )
    logging.info(f"Here is the model:\n{model}")

    # Get the tester and run!
    tester = ModuleEvaluator(opt, model, test_loader)
    tester.run()


if __name__ == "__main__":
    main()
