#!/usr/bin/env python

import io as IO
import json
import logging
import os

import matplotlib.pyplot as plt
import redis
import torch
from flask import Flask, jsonify, request, send_file, session
from flask_session import Session
from skimage import io

import nmn.server.globals as globals
from nmn.datasets.load.utils.transforms import ImageToTensor
from nmn.options.get_options import get_options
from nmn.server.exceptions import MissingCaseException, MissingKeyException
from nmn.server.primitives import (
    Context,
    Count,
    Exist,
    Filter,
    FindInContext,
    ImmediateRelate,
    Intersect,
    Query,
    Relate,
    Same,
    SetDifference,
    Union,
    Unique,
)
from nmn.server.utils import check_request_data, clear_all_attentions, get_attention, load_modules, same_attention
from nmn.utils.config import load_experiment_config

# app setup
app = Flask(__name__)
app.secret_key = b"0WppMdTQiaS47HUD76wrjA=="  # This is a secret
app.config["SESSION_TYPE"] = "redis"
app.config["SESSION_PERMANENT"] = True
app.config["SESSION_USE_SIGNER"] = True
app.config["SESSION_REDIS"] = redis.from_url("redis://localhost:6379")
app.config["SESSION_FILE_THRESHOLD"] = 1000
server_session = Session(app)


# Primitive routes
@app.route("/get-context", methods=["POST"])
def get_context():
    request_data = request.get_json(force=True)
    primitive = Context(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/find-in-context", methods=["POST"])
def find_in_context():
    request_data = request.get_json(force=True)
    primitive = FindInContext(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/set-difference", methods=["POST"])
def set_difference():
    request_data = request.get_json(force=True)
    primitive = SetDifference(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/count", methods=["POST"])
def count():
    request_data = request.get_json(force=True)
    primitive = Count(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/exist", methods=["POST"])
def exist():
    request_data = request.get_json(force=True)
    primitive = Exist(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/filter", methods=["POST"])
def filter_primitive():  # filter is a built-in
    request_data = request.get_json(force=True)
    primitive = Filter(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/intersect", methods=["POST"])
def intersect():
    request_data = request.get_json(force=True)
    primitive = Intersect(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/query", methods=["POST"])
def query():
    request_data = request.get_json(force=True)
    primitive = Query(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/relate", methods=["POST"])
def relate():
    request_data = request.get_json(force=True)
    primitive = Relate(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/same", methods=["POST"])
def same():
    request_data = request.get_json(force=True)
    primitive = Same(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/union", methods=["POST"])
def union():
    request_data = request.get_json(force=True)
    primitive = Union(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/unique", methods=["POST"])
def unique():
    request_data = request.get_json(force=True)
    primitive = Unique(request_data, globals.args, globals.modules)
    return primitive.evaluate()


@app.route("/immediate-relate", methods=["POST"])
def immediate_relate():
    request_data = request.get_json(force=True)
    primitive = ImmediateRelate(request_data, globals.args, globals.modules)
    return primitive.evaluate()


# Other routes
@app.route("/init-image", methods=["POST"])
def init_image():
    """
    The init_image route should be called before the start
    of the execution of an IRL program. This route
    prepares the server for executing primitives over this
    scene/image. The route expects the name of an image.
    This image is loaded and stored as a torch tensor.
    This allows subsequent primitives to easily use it.
    Additionally, all previously stored attentions are removed.
    These could be left over from previous executions. This
    will prevent that the server's memory explodes.
    """
    request_data = request.get_json(force=True)
    error_msg = check_request_data(request_data, ["name"])
    if error_msg is not None:
        raise MissingKeyException(error_msg)
    else:
        # clear the attentions currently stored in the session
        clear_all_attentions(session)
        # load the image into the session
        if "stored_images" not in session:
            session["stored_images"] = {}
        img_name = request_data["name"]
        scene_path = os.path.join(globals.args.images_dir, img_name)
        if img_name in session["stored_images"]:
            # restore a previously loaded image
            session["current_scene_path"] = scene_path
            session["current_scene"] = session["stored_images"][img_name]
        else:
            # load an image from file, transform it to a tensor
            # and store it in the session
            session["current_scene_path"] = scene_path
            image = io.imread(scene_path)
            img2tensor = ImageToTensor("cpu")
            tensor = img2tensor(image)
            tensor = torch.stack([tensor])
            session["current_scene"] = tensor
            session["stored_images"][img_name] = tensor
        return json.dumps({}), 200


@app.route("/clear-session", methods=["POST"])
def clear_session():
    """
    This invalidates the cookie and sends it back
    such that it can be deleted on the Lisp side.
    Also, it is removed from the redis db.
    """
    session.clear()
    return json.dumps({}), 200


@app.route("/clear-attentions", methods=["POST"])
def clear_attentions():
    """This removes all attentions from the redis db"""
    clear_all_attentions(session)
    return json.dumps({}), 200


@app.route("/attn/<attn_id>", methods=["GET"])
def get_attn(attn_id):
    # Get the image
    image = io.imread(session["current_scene_path"])
    # Get the attention
    attn = get_attention(attn_id, session)
    attn = attn.reshape((globals.args.config.img_height, globals.args.config.img_width)).detach().numpy()
    # Overlay them
    fig = plt.figure(frameon=False)
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(image, cmap="gray")
    ax.imshow(attn, cmap="magma", alpha=0.5)
    plt.axis("off")
    # Export
    out = IO.BytesIO()
    fig.savefig(out, bbox_inches="tight")
    out.seek(0)
    plt.close(fig=fig)
    return send_file(out, mimetype="image/png")


@app.route("/equal-entity", methods=["POST"])
def equal_entity():
    request_data = request.get_json(force=True)
    error_msg = check_request_data(request_data, ["attn1", "attn2"])
    if error_msg is not None:
        raise MissingKeyException(error_msg)
    else:
        attn_id_1 = request_data["attn1"]
        attn_id_2 = request_data["attn2"]
        attn_1 = get_attention(attn_id_1)
        attn_2 = get_attention(attn_id_2)
        equal_entity_p = same_attention(attn_1, attn_2, globals.args.iou_threshold)
        return json.dumps({"equalp": equal_entity_p}), 200


# Error handlers
@app.errorhandler(MissingKeyException)
def missing_key(err):
    return jsonify(err.to_dict())


@app.errorhandler(MissingCaseException)
def unknown_case(err):
    return jsonify(err.to_dict())


# Server start up
# This code is outside of __name__ == __main__
# because the Penelope server uses a different
# procedure to start the server.
globals.args = get_options("server")


# When running on the cluster, adjust
# the number of processes spawned by
# PyTorch
if "PBS_NP" in os.environ:
    num_torch_cores = globals.args.num_torch_processes
    torch.set_num_threads(num_torch_cores)
    torch.set_num_interop_threads(1)
    logging.info(f"PyTorch can use {num_torch_cores} cores")

# Turn off the logging (it leads to
# extremely large files on the cluster)
if globals.args.use_logging == 0:
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

load_experiment_config(globals.args)
load_modules(globals.args)
plt.switch_backend("Agg")

if __name__ == "__main__":
    app.run(
        host=globals.args.server_host,
        port=globals.args.server_port,
        threaded=False,
        processes=globals.args.num_flask_processes,
    )
