#!/usr/bin/env python

"""
Neural Module Server
This server exposes an API to access all modules.
It can be used to integrate the neural modules in IRL.
"""

import argparse
import io as IO
import json
import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import redis
import torch
from flask import Flask, request, send_file, session
from flask_session import Session
from skimage import io
from sklearn.metrics import jaccard_score

from nmn.datasets.experiment import Experiment
from nmn.datasets.load.utils.transforms import ImageToTensor
from nmn.modules.load_modules.get_modules import get_evaluation_modules
from nmn.utils.config import load_experiment_config

# Parsing arguments for server start up
#######################################
parser = argparse.ArgumentParser()
parser.add_argument(
    "--experiment",
    required=True,
    choices=Experiment.list(),
    help="Specify the experiment you are running",
)
parser.add_argument("--images-dir", required=True, help="Path to the images folder")
parser.add_argument("--server-host", default="127.0.0.1")
parser.add_argument("--server-port", default=8888, type=int)
parser.add_argument(
    "--specificity", default="concept", type=str, choices=["category", "concept"]
)
parser.add_argument(
    "--iou-threshold",
    default=0.9,
    type=float,
    help="Threshold for when attentions are considered equal",
)
parser.add_argument(
    "--num-flask-processes",
    default=1,
    type=int,
    help="Number of processes to spawn for handling requests",
)
parser.add_argument(
    "--num-torch-processes",
    default=1,
    type=int,
    help="Number of processes to spawn for PyTorch",
)
parser.add_argument(
    "--use-cuda", action="store_true", help="Whether to use cuda or not"
)
parser.add_argument(
    "--use-logging",
    default=1,
    type=int,
    choices=[0, 1],
    help="Logging can be turned off as it leads to huge files",
)

# app setup
###########
app = Flask(__name__)
app.secret_key = b"0WppMdTQiaS47HUD76wrjA=="  # This is a secret
app.config["SESSION_TYPE"] = "redis"
app.config["SESSION_PERMANENT"] = True
app.config["SESSION_USE_SIGNER"] = True
app.config["SESSION_REDIS"] = redis.from_url("redis://localhost:6379")
app.config["SESSION_FILE_THRESHOLD"] = 1000
server_session = Session(app)

# Global variables
##################
ARGS = None
MODULES = None
ATTENTION_COUNTER = 0


# init_image route
##################
@app.route("/init_image", methods=["POST"])
def init_image():
    """
    The init_image route should be called before the start
    of the execution of an IRL program. This route will
    prepare the server for executing primitives over this
    scene/image. The route expects the name of an image.
    This image will be loaded and stored as a torch tensor.
    This will allow subsequent primitives to easily use it.
    Additionally, all previously stored attentions will be removed.
    These could be left over from previous executions. This
    will prevent that the server's memory explodes.
    """
    request_data = request.get_json(force=True)
    errors = check_request_data(request_data, ["name"])
    if errors is not None:
        return errors, 400
    else:
        # clear the attentions currently stored in the session
        clear_all_attentions(session)
        # load the image into the session
        load_image(request_data["name"], session)
        return json.dumps({}), 200


# clear_session route
#####################
@app.route("/clear_session", methods=["POST"])
def clear_session():
    """
    This invalidates the cookie and sends it back
    such that it can be deleted on the Lisp side.
    Also, it is removed from the redis db.
    """
    session.clear()
    return json.dumps({}), 200


# clear_attentions route
#####################
@app.route("/clear_attentions", methods=["POST"])
def clear_attentions():
    """
    This removes all attentions from the redis db
    """
    clear_all_attentions(session)
    return json.dumps({}), 200


# get attention visualisation
#############################
@app.route("/attn/<attn_id>", methods=["GET"])
def get_attn(attn_id):
    # Get the image
    image = io.imread(session["current_scene_path"])
    # Get the attention
    global ARGS
    attn = get_attention(attn_id, session)
    attn = (
        attn.reshape((ARGS.config.img_height, ARGS.config.img_width)).detach().numpy()
    )
    # Overlay them
    fig = plt.figure(frameon=False)
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(image, cmap="gray")
    ax.imshow(attn, cmap="magma", alpha=0.5)
    plt.axis("off")
    # Export
    out = IO.BytesIO()
    fig.savefig(out, bbox_inches="tight")
    out.seek(0)
    plt.close(fig=fig)
    return send_file(out, mimetype="image/png")


# equal entity
##############
@app.route("/equal_entity", methods=["POST"])
def equal_entity():
    request_data = request.get_json(force=True)
    errors = check_request_data(request_data, ["attn1", "attn2"])
    if errors is not None:
        return errors, 400
    else:
        attn_id_1 = request_data["attn1"]
        attn_id_2 = request_data["attn2"]
        attn_1 = get_attention(attn_id_1)
        attn_2 = get_attention(attn_id_2)
        equal_entity_p = same_attention(attn_1, attn_2)
        return json.dumps({"equalp": equal_entity_p}), 200


# evaluate_neural_primitive route
#################################
@app.route("/evaluate_neural_primitive", methods=["POST"])
def evaluate_neural_primitive():
    """
    The evaluate_neural_primitive route is responsible for
    the neural execution of an IRL primitive.
    The request expects the name of the primitive to evaluate
    and the slots of the primitive in the form of a dictionary.
    Some of the slots can be bound (any value apart from None)
    while others can be unbound (= None). The server should
    then figure out in which direction the primitive needs to
    be executed (i.e. fill in the unbound slot(s)) and return
    zero, one or multiple new bindings for the unbound slot(s).
    """
    request_data = request.get_json(force=True)
    errors = check_request_data(request_data, ["primitive", "slots"])
    if errors is not None:
        return errors, 400
    else:
        primitive = request_data["primitive"]
        slots = request_data["slots"]
        for key, value in slots.items():
            if value == []:
                slots[key] = None
        bound_slots = [key for key, value in slots.items() if value is not None]
        unbound_slots = [key for key, value in slots.items() if value is None]
        logging.info(f"{primitive} ({bound_slots} => {unbound_slots})")
        if primitive == "get_context":
            return get_context(slots)
        elif primitive == "count":
            return count(slots)
        elif primitive == "exist":
            return exist(slots)
        elif primitive == "filter":
            return filter_primitive(slots)
        elif primitive == "intersect":
            return intersect(slots)
        elif primitive == "query":
            return query(slots)
        elif primitive == "relate":
            return relate(slots)
        elif primitive == "same":
            return same(slots)
        elif primitive == "union":
            return union(slots)
        elif primitive == "unique":
            return unique(slots)
        # mnist concept primitives
        elif primitive == "immediate_relate":
            return immediate_relate(slots)
        elif primitive == "set_diff":
            return set_diff(slots)
        elif primitive == "find_in_context":
            return find_in_context(slots)
        else:
            return (
                json.dumps(
                    {
                        "error_type": "unknown_primitive",
                        "message": "Unknown primitive %s" % primitive,
                    }
                ),
                400,
            )


def get_context(slots):
    """
    The get_context module expects a single slot 'context'.
    (=> context)
    If this slot is unbound, execute the context module and bind
    the resulting attention to 'context'.
    (context =>)
    If this slot is bound, it must be some attention. Execute the
    context module and compare the result to the stored attention.
    """
    global ARGS
    # check the slots
    if "context" not in slots:
        return missing_slot_error("context", "get_context")
    # case 1
    if check_case(slots, unbound=["context"]):
        context_module = get_module("context")
        attn = context_module({"image": get_image(session)})
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "context", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(slots, bound=["context"]):
        context_module = get_module("context")
        stored_attention = get_attention(slots["context"], session)
        if ARGS.experiment == Experiment.MNIST_DIALOG:
            attn = get_image()
        else:
            attn = context_module({"image": get_image()})
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("context", slots)


def find_in_context(slots):
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "find_in_context")
    if "history" not in slots:
        return missing_slot_error("history", "find_in_context")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "find_in_context")
    # case 1
    if check_case(slots, unbound=["target_attn"]):
        # todo: add multiple history objects
        if (len(slots["history"])) == 1:
            attn_id = slots["history"][0]
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )


def set_diff(slots):
    """
    Set_diff takes 3 slots: 'source_attn_1', 'source_attn_2'
    and 'target_attn'.
    (source_attn_1 source_attn_2 => target_attn)
    Given source attentions, compute the set-diff as target attn
    (source_attn_1 source_attn_2 target_attn =>)
    Compute the set-diff of the source attentions and compare
    against target attn
    """
    # check slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "set_diff")
    if "history" not in slots:
        return missing_slot_error("history", "set_diff")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "set_diff")
    # case 1
    if check_case(slots, bound=["source_attn", "history"], unbound=["target_attn"]):
        set_diff_module = get_module("set_diff")
        combined_attn = 0
        for el in slots["history"]:
            att = get_attention(el, session)
            combined_attn = torch.add(combined_attn, att, alpha=1)

        source_attn = get_attention(slots["source_attn"], session)
        attn = set_diff_module(source_attn, combined_attn)
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    elif check_case(slots, bound=["source_attn", "history", "target_attn"]):
        set_diff_module = get_module("set_diff")
        source_attn = get_attention(slots["source_attn"], session)
        history = get_attention(slots["history"], session)
        attn = set_diff_module(source_attn, history)
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": consistent}), 200

    else:
        return unknown_case_error("set_diff", slots)


def count(slots):
    """
    The count module gets two slots: 'source_attn' and 'target_num'.
    (source_attn => target_num)
    Run the count module on the attn and return a new binding
    for the target (= an integer)
    (source_attn target_num =>)
    Run the count module on the attn and compare the result
    against target
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "count")
    if "target_num" not in slots:
        return missing_slot_error("target_num", "count")
    # case 1
    if check_case(slots, bound=["source_attn"], unbound=["target_num"]):
        count_module = get_module("count")
        log_probs = count_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
            }
        )
        max_log_prob, n = torch.max(log_probs, 1)
        max_prob = np.exp(max_log_prob.item())
        n = n.item()
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_num", "score": max_prob, "value": n}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(slots, bound=["source_attn", "target_num"]):
        count_module = get_module("count")
        log_probs = count_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
            }
        )
        n = log_probs.argmax(dim=1).item()
        consistent = n == slots["target_num"]
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("count", slots)


def exist(slots):
    """
    The exist module takes 2 slots: 'source_attn' and 'target_bool'.
    (source_attn => target_bool)
    Run the exist module on the source attn and return a new binding
    for target bool
    (source_attn target_bool =>)
    Run the exist module on the source attn and compare against
    target bool
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "exist")
    if "target_bool" not in slots:
        return missing_slot_error("target_bool", "exist")
    # case 1
    if check_case(slots, bound=["source_attn"], unbound=["target_bool"]):
        exist_module = get_module("exist")
        bool, prob = exist_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
            }
        )
        target = "yes" if bool else "no"
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_bool", "score": prob, "value": target}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(slots, bound=["source_attn", "target_bool"]):
        exist_module = get_module("exist")
        bool, _ = exist_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
            }
        )
        target = "yes" if bool else "no"
        consistent = target == slots["target_bool"]
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("exist", slots)


def filter_primitive(slots):
    """
    The filter primitive takes 3 slots: 'source_attn',
    'target_attn' and 'category'.
    (source_attn category => target_attn)
    Run the filter module and bind the target attn
    (source_attn target_attn => category)
    Try the filter module with all categories and compare
    against the target attn
    (source_attn => target_attn category)
    Run the filter module with all categories and create
    many bindings
    (source_attn target_attn category =>)
    Run the filter module with the category and compare
    against the target attn
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "filter")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "filter")
    if "category" not in slots:
        return missing_slot_error("category", "filter")
    global ARGS
    # case 1
    if check_case(slots, bound=["source_attn", "category"], unbound=["target_attn"]):
        filter_module = get_module("filter")
        # the module needs to have both the 'category' (e.g. color)
        # and the 'value' (e.g. blue). IRL always uses the latter,
        # so we have to retrieve the former
        filter_category = None
        for category, concepts in ARGS.config.categories.items():
            if str(slots["category"]) in concepts:
                filter_category = category
                break
        # !!!
        if slots["category"] == "thing":
            # filter_thing in CLEVR is just returning the same attention
            attn = get_attention(slots["source_attn"], session)
        elif slots["category"] == "number":
            # filter_number in MNIST is just returning the same attention
            attn = get_attention(slots["source_attn"], session)
        else:
            attn = filter_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "category": filter_category,
                    "value": str(slots["category"]),
                }
            )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(slots, bound=["source_attn", "target_attn"], unbound=["category"]):
        filter_module = get_module("filter")
        possible_values = []
        stored_attention = get_attention(slots["target_attn"], session)
        for category, concepts in ARGS.config.categories.items():
            for value in concepts:
                attn = filter_module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(slots["source_attn"], session),
                        "category": category,
                        "value": value,
                    }
                )
                iou = compute_iou(stored_attention, attn)
                if iou > ARGS.iou_threshold:
                    possible_values.append((value, iou))
        # here we choose to return a single value
        # other possibilty would be to return all concepts
        # that have iou > threshold, and use the iou
        # as score in IRL
        best_value, best_iou = max(possible_values, key=lambda p: p[1])
        return (
            json.dumps(
                {
                    "bindings": [
                        [
                            {
                                "variable": "category",
                                "score": best_iou,
                                "value": best_value,
                            }
                        ]
                    ]
                }
            ),
            200,
        )
    # case 3
    elif check_case(slots, bound=["source_attn"], unbound=["target_attn", "category"]):
        filter_module = get_module("filter")
        bindings = []
        # run the find module for all categories and concepts
        # and return all bindings
        for category, concepts in ARGS.config.categories.items():
            for value in concepts:
                attn = filter_module(
                    {
                        "image": get_image(session),
                        "mask": get_attention(slots["source_attn"], session),
                        "category": category,
                        "value": value,
                    }
                )
                attn_id = store_attention(attn, session)
                bindings.append(
                    [
                        {"variable": "target_attn", "score": 1.0, "value": attn_id},
                        {"variable": "category", "score": 1.0, "value": value},
                    ]
                )
        return json.dumps({"bindings": bindings}), 200
    # case 4
    elif check_case(slots, bound=["source_attn", "target_attn", "category"]):
        filter_module = get_module("filter")
        for category, concepts in ARGS.config.categories.items():
            if slots["category"] in concepts:
                filter_category = category
                break

        attn = filter_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "category": filter_category,
                "value": slots["category"],
            }
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("filter", slots)


def intersect(slots):
    """
    Intersect takes 3 slots: 'source_attn_1', 'source_attn_2'
    and 'target_attn'.
    (source_attn_1 source_attn_2 => target_attn)
    Given source attentions, compute the intersection as target attn
    (source_attn_1 source_attn_2 target_attn =>)
    Compute the intersection of the source attentions and compare
    against target attn
    """
    # check slots
    if "source_attn_1" not in slots:
        return missing_slot_error("source_attn_1", "intersect")
    if "source_attn_2" not in slots:
        return missing_slot_error("source_attn_2", "intersect")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "intersect")
    # case 1
    if check_case(
        slots, bound=["source_attn_1", "source_attn_2"], unbound=["target_attn"]
    ):
        intersect_module = get_module("intersect")
        attn = intersect_module(
            get_attention(slots["source_attn_1"], session),
            get_attention(slots["source_attn_2"], session),
        )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    elif check_case(slots, bound=["source_attn_1", "source_attn_2", "target_attn"]):
        intersect_module = get_module("intersect")
        attn = intersect_module(
            get_attention(slots["source_attn_1"], session),
            get_attention(slots["source_attn_2"], session),
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("intersect", slots)


def query(slots):
    """
    Query takes 3 slots: 'source_attn', 'attribute' and
    'target_category'.
    (source_attn attribute => target_category)
    (source_attn target_category => attribute)
    (source_attn => attribute target_category)
    (source_attn attribute target_category =>)
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "query")
    if "attribute" not in slots:
        return missing_slot_error("attribute", "query")
    if "target_category" not in slots:
        return missing_slot_error("target_category", "query")
    global ARGS
    # case 1
    if check_case(
        slots, bound=["source_attn", "attribute"], unbound=["target_category"]
    ):
        query_module = get_module("query")
        options = ARGS.config.categories[slots["attribute"]]
        index, prob = query_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "category": slots["attribute"],
                "options": options,
            }
        )
        label = ARGS.config.categories[slots["attribute"]][index]
        if ARGS.experiment == Experiment.MNIST_DIALOG:
            label = convert_int_to_number(label)
            if slots["attribute"] == "bgcolor":
                label = label + "-bg"
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_category", "score": prob, "value": label}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(
        slots, bound=["source_attn", "target_category"], unbound=["attribute"]
    ):
        query_module = get_module("query")
        bindings = []
        for category, concepts in ARGS.config.categories.items():
            index, prob = query_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "category": category,
                    "options": concepts,
                }
            )
            label = ARGS.config.categories[category][index]
            if label == slots["target_category"]:
                bindings.append(
                    [{"variable": "attribute", "score": prob, "value": category}]
                )
        return json.dumps({"bindings": bindings}), 200
    # case 3
    elif check_case(
        slots, bound=["source_attn"], unbound=["attribute", "target_category"]
    ):
        query_module = get_module("query")
        bindings = []
        for category, concepts in ARGS.config.categories.items():
            index, prob = query_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "category": category,
                    "options": concepts,
                }
            )
            label = ARGS.config.categories[category][index]
            bindings.append(
                [
                    {"variable": "attribute", "score": 1.0, "value": category},
                    {"variable": "target_category", "score": prob, "value": label},
                ]
            )
        return json.dumps({"bindings": bindings}), 200
    # case 4
    elif check_case(slots, bound=["source_attn", "attribute", "target_category"]):
        query_module = get_module("query")
        options = ARGS.config.categories[slots["attribute"]]
        index, _ = query_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "category": slots["attribute"],
                "options": options,
            }
        )
        label = ARGS.config.categories[slots["attribute"]][index]
        consistent = label == slots["target_category"]
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("query", slots)


def relate(slots):
    """
    Relate takes 3 slots: 'source_attn', 'spatial_relation'
    and 'target_attn'.
    (source_attn spatial_relation => target_attn)
    (source_attn target_attn => spatial_relation)
    (source_attn => spatial_relation target_attn)
    (source_attn spatial_relation target_attn =>)
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "relate")
    if "spatial_relation" not in slots:
        return missing_slot_error("spatial_relation", "relate")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "relate")
    global ARGS
    # case 1
    if check_case(
        slots, bound=["source_attn", "spatial_relation"], unbound=["target_attn"]
    ):
        relate_module = get_module("relate")
        attn = relate_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "relationship": slots["spatial_relation"],
            }
        )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(
        slots, bound=["source_attn", "target_attn"], unbound=["spatial_relation"]
    ):
        relate_module = get_module("relate")
        possible_relations = []
        stored_attention = get_attention(slots["target_attn"], session)
        for relation in ARGS.config.relate:
            attn = relate_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "relationship": relation,
                }
            )
            iou = compute_iou(stored_attention, attn)
            if iou > ARGS.iou_threshold:
                possible_relations.append((relation, iou))
        best_relation, best_iou = max(possible_relations, key=lambda p: p[1])
        return (
            json.dumps(
                {
                    "bindings": [
                        [
                            {
                                "variable": "spatial_relation",
                                "score": best_iou,
                                "value": best_relation,
                            }
                        ]
                    ]
                }
            ),
            200,
        )
    # case 3
    elif check_case(
        slots, bound=["source_attn"], unbound=["spatial_relation", "target_attn"]
    ):
        relate_module = get_module("relate")
        bindings = []
        for relation in ARGS.config.relate:
            attn = relate_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "relationship": relation,
                }
            )
            attn_id = store_attention(attn, session)
            bindings.append(
                [
                    {"variable": "spatial_relation", "score": 1.0, "value": relation},
                    {"variable": "target_attn", "score": 1.0, "value": attn_id},
                ]
            )
        return json.dumps({"bindings": bindings}), 200
    # case 4
    elif check_case(slots, bound=["source_attn", "spatial_relation", "target_attn"]):
        relate_module = get_module("relate")
        attn = relate_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "relationship": slots["spatial_relation"],
            }
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("relate", slots)


def same(slots):
    """
    Same takes 3 slots: 'source_attn', 'attribute'
    and 'target_attn'
    (source_attn attribute => target_attn)
    (source_attn target_attn => attribute)
    (source_attn => attribute target_attn)
    (source_attn attribute target_attn =>)
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "same")
    if "attribute" not in slots:
        return missing_slot_error("attribute", "same")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "same")
    global ARGS
    # case 1
    if check_case(slots, bound=["source_attn", "attribute"], unbound=["target_attn"]):
        same_module = get_module("same")
        options = ARGS.config.categories[slots["attribute"]]
        attn = same_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "category": slots["attribute"],
                "options": options,
            }
        )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(slots, bound=["source_attn", "target_attn"], unbound=["attribute"]):
        same_module = get_module("same")
        possible_values = []
        stored_attention = get_attention(slots["target_attn"], session)
        for category, options in ARGS.config.categories.items():
            attn = same_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "category": category,
                    "options": options,
                }
            )
            iou = compute_iou(stored_attention, attn)
            if iou > ARGS.iou_threshold:
                possible_values.append((category, iou))
        best_value, best_iou = max(possible_values, key=lambda p: p[1])
        return (
            json.dumps(
                {
                    "bindings": [
                        [
                            {
                                "variable": "attribute",
                                "score": best_iou,
                                "value": best_value,
                            }
                        ]
                    ]
                }
            ),
            200,
        )
    # case 3
    elif check_case(slots, bound=["source_attn"], unbound=["attribute", "target_attn"]):
        same_module = get_module("same")
        bindings = []
        for category, options in ARGS.config.categories.items():
            attn = same_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "category": category,
                    "options": options,
                }
            )
            attn_id = store_attention(attn, session)
            bindings.append(
                [
                    {"variable": "attribute", "score": 1.0, "value": category},
                    {"variable": "target_attn", "score": 1.0, "value": attn_id},
                ]
            )
        return json.dumps({"bindings": bindings}), 200
    # case 4
    elif check_case("slots", bound=["source_attn", "attribute", "target_attn"]):
        same_module = get_module("same")
        options = ARGS.config.categories[slots["attribute"]]
        attn = same_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "category": slots["attribute"],
                "options": options,
            }
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("same", slots)


def union(slots):
    """
    Union takes 3 slots: 'source_attn_1', 'source_attn_2'
    and 'target_attn'.
    (source_attn_1 source_attn_2 => target_attn)
    Given source attentions, compute the union as target attn
    (source_attn_1 source_attn_2 target_attn =>)
    This requires to compare attentions
    """
    # check slots
    if "source_attn_1" not in slots:
        return missing_slot_error("source_attn_1", "union")
    if "source_attn_2" not in slots:
        return missing_slot_error("source_attn_2", "union")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "union")
    # case 1
    if check_case(
        slots, bound=["source_attn_1", "source_attn_2"], unbound=["target_attn"]
    ):
        union_module = get_module("union")
        attn = union_module(
            get_attention(slots["source_attn_1"], session),
            get_attention(slots["source_attn_2"], session),
        )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    elif check_case(slots, bound=["source_attn_1", "source_attn_2", "target_attn"]):
        union_module = get_module("union")
        attn = union_module(
            get_attention(slots["source_attn_1"], session),
            get_attention(slots["source_attn_2"], session),
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("union", slots)


def unique(slots):
    """
    Unique has 2 slots: 'source_attn' and 'target_attn'
    (source_attn => target_attn)
    (source_attn target_attn =>)
    """
    # check slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "unique")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "unique")
    # case 1
    if check_case(slots, bound=["source_attn"], unbound=["target_attn"]):
        unique_module = get_module("unique")
        att = get_attention(slots["source_attn"], session)
        bool, prob = unique_module({"image": get_image(session), "mask": att})
        if bool:
            # if unique, return the source attn as target attn
            return (
                json.dumps(
                    {
                        "bindings": [
                            [
                                {
                                    "variable": "target_attn",
                                    "score": prob,
                                    "value": slots["source_attn"],
                                }
                            ]
                        ]
                    }
                ),
                200,
            )
        else:
            # return no bindings, forcing IRL to stop
            return json.dumps({"bindings": []}), 200
    # case 2
    elif check_case(slots, bound=["source_attn", "target_attn"]):
        unique_module = get_module("unique")
        attn = get_attention(slots["source_attn"], session)
        bool, _ = unique_module({"image": get_image(session), "mask": attn})
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": bool and consistent}), 200
    else:
        return unknown_case_error("unique", slots)


def immediate_relate(slots):
    """
    Relate takes 3 slots: 'source_attn', 'spatial_relation'
    and 'target_attn'.
    (source_attn spatial_relation => target_attn)
    (source_attn target_attn => spatial_relation)
    (source_attn => spatial_relation target_attn)
    (source_attn spatial_relation target_attn =>)
    """
    # check the slots
    if "source_attn" not in slots:
        return missing_slot_error("source_attn", "immediate_relate")
    if "spatial_relation" not in slots:
        return missing_slot_error("spatial_relation", "immediate_relate")
    if "target_attn" not in slots:
        return missing_slot_error("target_attn", "immediate_relate")
    global ARGS
    # case 1
    if check_case(
        slots, bound=["source_attn", "spatial_relation"], unbound=["target_attn"]
    ):
        relate_module = get_module("immediate_relate")
        attn = relate_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "relationship": slots["spatial_relation"],
            }
        )
        attn_id = store_attention(attn, session)
        return (
            json.dumps(
                {
                    "bindings": [
                        [{"variable": "target_attn", "score": 1.0, "value": attn_id}]
                    ]
                }
            ),
            200,
        )
    # case 2
    elif check_case(
        slots, bound=["source_attn", "target_attn"], unbound=["spatial_relation"]
    ):
        relate_module = get_module("immediate_relate")
        possible_relations = []
        stored_attention = get_attention(slots["target_attn"], session)
        for relation in ARGS.config.relate:
            attn = relate_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "relationship": relation,
                }
            )
            iou = compute_iou(stored_attention, attn)
            if iou > ARGS.iou_threshold:
                possible_relations.append((relation, iou))
        best_relation, best_iou = max(possible_relations, key=lambda p: p[1])
        return (
            json.dumps(
                {
                    "bindings": [
                        [
                            {
                                "variable": "spatial_relation",
                                "score": best_iou,
                                "value": best_relation,
                            }
                        ]
                    ]
                }
            ),
            200,
        )
    # case 3
    elif check_case(
        slots, bound=["source_attn"], unbound=["spatial_relation", "target_attn"]
    ):
        relate_module = get_module("immediate_relate")
        bindings = []
        for relation in ARGS.config.relate:
            attn = relate_module(
                {
                    "image": get_image(session),
                    "mask": get_attention(slots["source_attn"], session),
                    "relationship": relation,
                }
            )
            attn_id = store_attention(attn, session)
            bindings.append(
                [
                    {"variable": "spatial_relation", "score": 1.0, "value": relation},
                    {"variable": "target_attn", "score": 1.0, "value": attn_id},
                ]
            )
        return json.dumps({"bindings": bindings}), 200
    # case 4
    elif check_case(slots, bound=["source_attn", "spatial_relation", "target_attn"]):
        relate_module = get_module("immediate_relate")
        attn = relate_module(
            {
                "image": get_image(session),
                "mask": get_attention(slots["source_attn"], session),
                "relationship": slots["spatial_relation"],
            }
        )
        stored_attention = get_attention(slots["target_attn"], session)
        consistent = same_attention(stored_attention, attn)
        return json.dumps({"consistent": str(consistent)}), 200
    else:
        return unknown_case_error("immediate_relate", slots)


# Helper functions
##################


def check_case(slots, unbound=[], bound=[]):
    """
    Check if an IRL case holds
    """
    unbound_check = all([slots[slot] is None for slot in unbound])
    bound_check = all([slots[slot] is not None for slot in bound])
    return unbound_check and bound_check


def get_attention(attn_id, session):
    """
    Retrieve an attention by its ID
    """
    return session["attentions"][attn_id]


def same_attention(attn_1, attn_2):
    iou = compute_iou(attn_1, attn_2)
    # Check the threshold
    global ARGS
    return iou >= ARGS.iou_threshold


def compute_iou(attn_1, attn_2):
    # Discretize the attentions
    mask_1 = np.where(attn_1 <= 0.5, 0, 1)
    mask_2 = np.where(attn_2 <= 0.5, 0, 1)
    # Flatten to single dim
    mask_1 = mask_1.reshape(-1)
    mask_2 = mask_2.reshape(-1)
    # Compute IOU
    iou = jaccard_score(mask_1, mask_2, average="macro")
    return iou


def store_attention(attn, session):
    """
    Store a new attention
    """
    global ATTENTION_COUNTER
    ATTENTION_COUNTER += 1
    name = "attn_" + str(ATTENTION_COUNTER)
    session["attentions"][name] = attn
    return name


def get_image(session):
    """
    Retrieve the scene
    """
    return session["current_scene"]


def get_module(name):
    """
    Retrieve a module
    """
    global MODULES
    return MODULES[name]


def check_request_data(request_data, keys):
    """
    Helper function that checks if certain key(s) are present
    in the POST data and if they are, returns the data.
    """
    for key in keys:
        if key not in request_data:
            return json.dumps(
                {
                    "error_type": "missing_key",
                    "message": 'No data provided using the "{}" key.'.format(key),
                }
            )
    return None


def clear_all_attentions(session):
    """
    Clear all attentions stored in the session
    """
    if "attentions" in session:
        num_stored_attentions = len(session["attentions"])
        logging.info(f"Clearing {num_stored_attentions} attentions.")
    session["attentions"] = {}


def missing_slot_error(slot_name, primitive):
    """
    Formulate a missing slot error
    """
    return (
        json.dumps(
            {
                "error_type": "missing_slot",
                "message": f"Missing slot {slot_name} in primitive {primitive}",
            }
        ),
        400,
    )


def unknown_case_error(primitive, slots):
    """
    Formulate an unknown case error
    """
    bound_slots = [slot for slot in slots if slots[slot] is not None]
    unbound_slots = [slot for slot in slots if slots[slot] is None]
    case = f"{bound_slots} => {unbound_slots}"
    return (
        json.dumps(
            {
                "error_type": "unknown_case",
                "message": f"Unknown case for primitive {primitive}: {case}",
            }
        ),
        400,
    )


def load_image(name, session):
    global ARGS
    if "stored_images" in session:
        if name in session["stored_images"]:
            session["current_scene_path"] = os.path.join(ARGS.images_dir, name)
            session["current_scene"] = session["stored_images"][name]
            # logging.info(f'Restoring image {name}')
        else:
            scene_path = os.path.join(ARGS.images_dir, name)
            # logging.info(f'Loading fresh image from {scene_path}')
            session["current_scene_path"] = scene_path
            image = io.imread(scene_path)
            img2tensor = ImageToTensor("cpu")
            tensor = img2tensor(image)
            tensor = torch.stack([tensor])
            session["current_scene"] = tensor
            session["stored_images"][name] = tensor
    else:
        session["stored_images"] = {}
        load_image(name, session)


def load_modules(args):
    kwargs = {
        "specificity": args.specificity,
        "experiment": args.experiment,
        "config": args.config,
        "device": "cpu",
    }
    global MODULES
    MODULES = get_evaluation_modules(**kwargs)
    logging.info("Loaded the following modules:")
    logging.info(MODULES.keys())


def convert_int_to_number(num):
    if num == "0":
        new_num = "zero"
    elif num == "1":
        new_num = "one"
    elif num == "2":
        new_num = "two"
    elif num == "3":
        new_num = "three"
    elif num == "4":
        new_num = "four"
    elif num == "5":
        new_num = "five"
    elif num == "6":
        new_num = "six"
    elif num == "7":
        new_num = "seven"
    elif num == "8":
        new_num = "eight"
    elif num == "9":
        new_num = "nine"
    else:
        new_num = num
    return new_num


# Server start up
#################
ARGS = parser.parse_args()

# Add device to args
ARGS.device = None
if ARGS.use_cuda and torch.cuda.is_available():
    logging.info("CUDA device available")
    ARGS.device = "cuda:0"
else:
    logging.info("No CUDA device available")
    ARGS.device = "cpu"

# When running on the cluster, adjust
# the number of processes spawned by
# PyTorch
if "PBS_NP" in os.environ:
    num_torch_cores = ARGS.num_torch_processes
    torch.set_num_threads(num_torch_cores)
    torch.set_num_interop_threads(1)
    logging.info(f"PyTorch can use {num_torch_cores} cores")

# Turn off the logging (it leads to
# extremely large files on the cluster)
if ARGS.use_logging == 0:
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

load_experiment_config(ARGS)
load_modules(ARGS)
plt.switch_backend("Agg")

if __name__ == "__main__":
    app.run(
        host=ARGS.server_host,
        port=ARGS.server_port,
        threaded=False,
        processes=ARGS.num_flask_processes,
    )
