#!/usr/bin/env python

"""
Use this file to train 
a Mask R-CNN or Faster R-CNN model
on the CLEVR mini dataset
using the Detectron2 framework
"""

import argparse
import json
import os

import cv2 as cv
import pycocotools.mask
from detectron2 import model_zoo
from detectron2.config import get_cfg
from detectron2.data import DatasetCatalog, MetadataCatalog
from detectron2.engine import DefaultTrainer
from detectron2.structures import BoxMode

from nmn.utils.logger import setup_logger

parser = argparse.ArgumentParser()
# Input/Output args
parser.add_argument(
    "--clevr_mini_images",
    type=str,
    default="data/clevr_mini/images",
    help="Path to CLEVR mini images",
)
parser.add_argument(
    "--clevr_mini_scenes",
    type=str,
    default="data/clevr_mini/CLEVR_mini_coco_anns.json",
    help="Path to CLEVR mini scenes with COCO annotations",
)
parser.add_argument(
    "--output_dir",
    type=str,
    default="models/clevr_mini/",
    help="Where to output the trained model",
)
# Parameters of the model
parser.add_argument("--batch_size", default=8, type=int)
parser.add_argument("--max_iterations", default=100000, type=int)
parser.add_argument("--learning_rate", default=0.00025, type=float)
parser.add_argument("--use_cuda", action="store_true")
# Which model to train
parser.add_argument(
    "--model", default="mask_rcnn", type=str, choices=["mask_rcnn", "faster_rcnn"]
)
parser.add_argument(
    "--backbone",
    default="R_50_FPN",
    type=str,
    choices=["R_50_FPN", "R_50_C4", "R_50_DC5", "R_101_FPN", "R_101_C5", "R_101_DC5"],
)
parser.add_argument("--schedule", default="1x", type=str, choices=["1x", "3x"])


def get_clevr_mini_dataset(images_dir, scenes_json):
    """
    Process the CLEVR mini dataset to obtain the representation
    required by Detectron. Each images requires:
    - a file name
    - a unique ID
    - width and height
    - annotations consisting of bbox, mask and label
    """
    shape_to_category = {"sphere": 0, "cylinder": 1, "cube": 2}
    dataset = []
    with open(scenes_json) as f:
        data = json.load(f)
    scenes = data["scenes"]
    for i, scene in enumerate(scenes):
        img_path = os.path.join(images_dir, scene["image_filename"])
        height, width = cv.imread(img_path).shape[:2]
        objs = []
        for o in scene["objects"]:
            obj = {
                "bbox": list(pycocotools.mask.toBbox([o["mask"]])[0]),
                "bbox_mode": BoxMode.XYWH_ABS,
                "segmentation": o["mask"],
                "category_id": shape_to_category[o["shape"]],
            }
            objs.append(obj)
        dataset.append(
            {
                "file_name": img_path,
                "image_id": i,
                "height": height,
                "width": width,
                "annotations": objs,
            }
        )
    return dataset


def get_detectron_path(args):
    """
    Parse the command line args to obtain a particular
    pretrained model from Detectron's model zoo
    """
    if args.model == "mask_rcnn":
        return (
            "_".join(
                ["COCO-InstanceSegmentation/mask_rcnn", args.backbone, args.schedule]
            )
            + ".yaml"
        )
    elif args.model == "faster_rcnn":
        return (
            "_".join(["COCO-Detection/faster_rcnn", args.backbone, args.schedule])
            + ".yaml"
        )


def get_model(args):
    """
    Build the configuration for the model, according
    to the command line arguments
    """
    detectron_path = get_detectron_path(args)
    cfg = get_cfg()
    cfg.merge_from_file(model_zoo.get_config_file(detectron_path))
    cfg.DATASETS.TRAIN = ("clevr_mini_train",)
    cfg.DATASETS.TEST = ()
    cfg.DATALOADER.NUM_WORKERS = 0
    cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(detectron_path)
    cfg.SOLVER.IMS_PER_BATCH = args.batch_size
    cfg.SOLVER.BASE_LR = args.learning_rate
    cfg.SOLVER.MAX_ITER = args.max_iterations
    cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = 128
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = args.n_classes
    cfg.MODEL.DEVICE = "cuda" if args.use_cuda else "cpu"
    cfg.INPUT.MASK_FORMAT = "bitmask"
    cfg.OUTPUT_DIR = args.output_dir
    return cfg


def main(args):
    # Register the CLEVR mini dataset
    DatasetCatalog.register(
        "clevr_mini_train",
        lambda: get_clevr_mini_dataset(args.clevr_mini_images, args.clevr_mini_scenes),
    )
    classes = [
        "sphere",
        "cylinder",
        "cube",
    ]
    args.n_classes = len(classes)
    MetadataCatalog.get("clevr_mini_train").thing_classes = classes
    # Get the configuration
    model_cfg = get_model(args)
    os.makedirs(model_cfg.OUTPUT_DIR, exist_ok=True)

    # log run
    setup_logger(args)

    # Make the trainer and train!
    trainer = DefaultTrainer(model_cfg)
    trainer.resume_or_load(resume=False)
    trainer.train()


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
