#!/usr/bin/env python

import logging

import torch

from nmn.datasets.load.dataloader import get_dataloader
from nmn.evaluate.evaluation_metrics import get_evaluation_fn
from nmn.modules.load_modules.get_module import get_module
from nmn.options.get_options import get_options
from nmn.train.loss import get_loss_fn
from nmn.train.model import Model
from nmn.train.trainer import Trainer
from nmn.utils.config import load_experiment_config, load_module_config
from nmn.utils.logger import setup_logger
from scripts.utils.loading import get_images, get_scenes


def main():
    # Parse command line args
    opt = get_options("train")

    # Read the config file
    load_experiment_config(opt)
    load_module_config(opt)

    # log run
    setup_logger(opt)

    # train the module!
    train(opt)


def train(opt):
    # Get train and val dataloaders
    train_loader_kwargs = {
        "device": opt.device,
        # dataloader options
        "num_workers": opt.num_workers,
        "shuffle": opt.shuffle_data and not opt.balance_classes,
        "batch_size": opt.batch_size,
        "balance_classes": opt.balance_classes,
        # train options
        "images": get_images(opt.experiment, "test"),
        "scenes": get_scenes(opt.experiment, "test", opt.module_name),
        "max_samples": opt.max_samples_train,
        # module options
        "module_info": opt.module_info,
        # config
        "config": opt.config,
    }
    train_loader = get_dataloader(**train_loader_kwargs)
    logging.info("train_loader has %d samples" % len(train_loader.dataset))
    val_loader_kwargs = {
        "device": opt.device,
        # dataloader options
        "num_workers": opt.num_workers,
        "shuffle": opt.shuffle_data and not opt.balance_classes,
        "batch_size": opt.batch_size,
        "balance_classes": opt.balance_classes,
        # train options
        "images": get_images(opt.experiment, "test"),
        "scenes": get_scenes(opt.experiment, "test", opt.module_name),
        "max_samples": opt.max_samples_val,
        # module options
        "module_info": opt.module_info,
        # config
        "config": opt.config,
    }
    val_loader = get_dataloader(**val_loader_kwargs)
    logging.info("val_loader has %d samples" % len(val_loader.dataset))

    # Get the module
    module_kwargs = opt.module_info
    module = get_module(opt.experiment, opt.device, **module_kwargs)
    checkpoint = None
    if opt.mode == "resume":
        checkpoint = torch.load(
            opt.checkpoint, map_location=lambda storage, loc: storage
        )
        state = checkpoint["state"]
        module.load_state_dict(state)

    # Get the loss function
    loss_fn = get_loss_fn(opt.module_info.type)

    # Get the evaluation function
    evaluation_fn = get_evaluation_fn(opt.module_info.type)

    # Use all of the above to create the model
    model = Model(
        module=module,
        loss_fn=loss_fn,
        evaluation_fn=evaluation_fn,
        learning_rate=opt.learning_rate,
        weight_decay=opt.weight_decay,
    )
    logging.info(f"Here is the model:\n{model}")

    # Get the trainer
    trainer = Trainer(
        opt, model, module_kwargs, train_loader, val_loader, checkpoint=checkpoint
    )
    # Run!
    trainer.run()


if __name__ == "__main__":
    main()
